import unittest

import numpy as np
from skfusion.fusion import Relation, ObjectType, \
    FusionGraph, Dfmc


class TestDfmc(unittest.TestCase):
    def test_dfmc(self):
        rnds = np.random.RandomState(0)
        R12 = rnds.rand(50, 30)

        t1 = ObjectType('type1', 50)
        t2 = ObjectType('type2', 30)
        relation = Relation(R12, t1, t2)
        fusion_graph = FusionGraph()
        fusion_graph.add_relation(relation)

        fuser = Dfmc(init_type='random', random_state=rnds).fuse(fusion_graph)
        self.assertEqual(fuser.backbone(relation).shape, (50, 30))
        self.assertEqual(fuser.factor(t1).shape, (50, 50))
        self.assertEqual(fuser.factor(t2).shape, (30, 30))
        np.testing.assert_almost_equal(fuser.complete(relation), relation.data)

    def test_masked(self):
        rnds = np.random.RandomState(0)
        R12 = np.ma.masked_less(rnds.rand(50, 30), 0.3)

        t1 = ObjectType('type1', 50)
        t2 = ObjectType('type2', 30)
        relation = Relation(R12, t1, t2)
        fusion_graph = FusionGraph()
        fusion_graph.add_relation(relation)

        fuser = Dfmc(init_type='random', random_state=rnds).fuse(fusion_graph)
        self.assertEqual(fuser.backbone(relation).shape, (50, 30))
        self.assertEqual(fuser.factor(t1).shape, (50, 50))
        self.assertEqual(fuser.factor(t2).shape, (30, 30))
        np.testing.assert_almost_equal(fuser.complete(relation)[~R12.mask], relation.data[~R12.mask])

    def test_preprocessors(self):
        rnds = np.random.RandomState(0)
        R12 = rnds.rand(50, 30)
        R12 = np.ma.masked_greater(R12, 0.7)

        t1 = ObjectType('type1', 50)
        t2 = ObjectType('type2', 30)

        def preprocessor(data):
            return np.ones_like(data)

        relation = Relation(R12, t1, t2, name='R', preprocessor=preprocessor)
        fusion_graph = FusionGraph()
        fusion_graph.add_relation(relation)

        fuser = Dfmc(init_type='random', random_state=rnds).fuse(fusion_graph)
        self.assertEqual(fuser.backbone(relation).shape, (50, 30))
        self.assertEqual(fuser.factor(t1).shape, (50, 50))
        self.assertEqual(fuser.factor(t2).shape, (30, 30))
        trnf = np.ones_like(relation.data)
        np.testing.assert_almost_equal(fuser.complete(relation), trnf)
        np.testing.assert_equal(fusion_graph.get_relation('R').data, R12)

    def test_postprocessors(self):
        rnds = np.random.RandomState(0)
        R12 = rnds.rand(50, 30)
        R12 = np.ma.masked_greater(R12, 0.7)

        t1 = ObjectType('type1', 50)
        t2 = ObjectType('type2', 30)

        def postprocessor(data):
            return data - 10

        relation = Relation(R12, t1, t2, name='R', postprocessor=postprocessor)
        fusion_graph = FusionGraph()
        fusion_graph.add_relation(relation)

        fuser = Dfmc(init_type='random', random_state=rnds).fuse(fusion_graph)
        self.assertEqual(fuser.backbone(relation).shape, (50, 30))
        self.assertEqual(fuser.factor(t1).shape, (50, 50))
        self.assertEqual(fuser.factor(t2).shape, (30, 30))
        trnf = relation.data - 10
        np.testing.assert_almost_equal(fuser.complete(relation), trnf)
        np.testing.assert_equal(fusion_graph.get_relation('R').data, R12)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from skfusion.fusion import Relation, ObjectType, \
    FusionGraph, Dfmf, Dfmc


class TestNRun(unittest.TestCase):
    def test_dfmf(self):
        rnds = np.random.RandomState(0)
        R12 = rnds.rand(30, 30)
        R13 = rnds.rand(30, 30)

        t1 = ObjectType('type1', 50)
        t2 = ObjectType('type2', 30)
        t3 = ObjectType('type3', 10)
        fusion_graph = FusionGraph()
        relations = [Relation(R12, t1, t2), Relation(R13, t1, t3)]
        fusion_graph.add_relations_from(relations)

        fuser = Dfmf(init_type='random', random_state=rnds, n_run=3
                     ).fuse(fusion_graph)
        self.assertEqual(len(list(fuser.factor(t1))), 3)
        self.assertEqual(len(list(fuser.factor(t2))), 3)
        self.assertEqual(len(list(fuser.factor(t3))), 3)
        self.assertEqual(len(list(fuser.backbone(relations[0]))), 3)
        self.assertEqual(len(list(fuser.backbone(relations[1]))), 3)
        for object_type in [t1, t2, t3]:
            for factor in fuser.factor(object_type):
                self.assertEqual(factor.shape, (30, object_type.rank))

        G1 = fuser.factor(t1, run=1)
        S13 = fuser.backbone(relations[1], run=1)
        G3 = fuser.factor(t3, run=1)
        R13_hat = np.dot(G1, np.dot(S13, G3.T))
        completed = fuser.complete(relations[1], run=1)
        np.testing.assert_almost_equal(completed, R13_hat)


    def test_dfmc(self):
        rnds = np.random.RandomState(0)
        R12 = rnds.rand(30, 30)
        R13 = rnds.rand(30, 30)

        t1 = ObjectType('type1', 50)
        t2 = ObjectType('type2', 30)
        t3 = ObjectType('type3', 10)
        fusion_graph = FusionGraph()
        relations = [Relation(R12, t1, t2), Relation(R13, t1, t3)]
        fusion_graph.add_relations_from(relations)

        fuser = Dfmc(init_type='random', random_state=rnds, n_run=3
                     ).fuse(fusion_graph)
        self.assertEqual(len(list(fuser.factor(t1))), 3)
        self.assertEqual(len(list(fuser.factor(t2))), 3)
        self.assertEqual(len(list(fuser.factor(t3))), 3)
        self.assertEqual(len(list(fuser.backbone(relations[0]))), 3)
        self.assertEqual(len(list(fuser.backbone(relations[1]))), 3)
        for object_type in [t1, t2, t3]:
            for factor in fuser.factor(object_type):
                self.assertEqual(factor.shape, (30, object_type.rank))

        G1 = fuser.factor(t1, run=1)
        S13 = fuser.backbone(relations[1], run=1)
        G3 = fuser.factor(t3, run=1)
        R13_hat = np.dot(G1, np.dot(S13, G3.T))
        completed = fuser.complete(relations[1], run=1)
        np.testing.assert_almost_equal(completed, R13_hat)


if __name__ == "__main__":
    unittest.main()



"""
==========================================================================
Fusion of three data sources for gene function prediction in Dictyostelium
==========================================================================

Fuse three data sets: gene expression data (Miranda et al., 2013, PLoS One),
slim gene annotations from Gene Ontology and protein-protein interaction
network from STRING database.

Learnt latent matrix factors are utilized for the prediction of slim GO
terms in Dictyostelium genes that are unavailable in the training phase.

This example demonstrates how latent matrices estimated by data fusion
can be utilized for association prediction.
"""
print(__doc__)

from sklearn import cross_validation, metrics
import numpy as np

from skfusion import datasets
from skfusion import fusion as skf


def main():
    n_folds = 10
    n_genes = dicty[gene][go_term][0].data.shape[0]
    cv = cross_validation.KFold(n_genes, n_folds=n_folds)
    fold_mse = np.zeros(n_folds)
    ann_mask = np.zeros_like(dicty[gene][go_term][0].data).astype('bool')

    relations = [
        skf.Relation(dicty[gene][go_term][0].data, gene, go_term),
        skf.Relation(dicty[gene][exp_cond][0].data, gene, exp_cond),
        skf.Relation(dicty[gene][gene][0].data, gene, gene)]
    fusion_graph = skf.FusionGraph(relations)
    fuser = skf.Dfmc(max_iter=30, n_run=1, init_type='random', random_state=0)

    for i, (train_idx, test_idx) in enumerate(cv):
        ann_mask[:] = False
        ann_mask[test_idx, :] = True
        fusion_graph[gene][go_term][0].mask = ann_mask

        fuser.fuse(fusion_graph)
        pred_ann = fuser.complete(fuser.fusion_graph[gene][go_term][0])[test_idx]
        true_ann = dicty[gene][go_term][0].data[test_idx]
        fold_mse[i] = metrics.mean_squared_error(pred_ann, true_ann)

    print("MSE: %5.4f" % np.mean(fold_mse))


if __name__ == "__main__":
    dicty = datasets.load_dicty()
    gene = dicty.get_object_type("Gene")
    go_term = dicty.get_object_type("GO term")
    exp_cond = dicty.get_object_type("Experimental condition")

    main()



"""
================================================================
Fusion of six data sources for pharmacological action prediction
================================================================

This example demonstrates how latent matrices estimated by data
fusion can be utilized for association prediction. The prediction
is done by random forest on features construction through reverse
chaining.
"""
print(__doc__)

from functools import reduce

from sklearn import cross_validation, metrics, ensemble
import numpy as np

from skfusion import datasets
from skfusion import fusion as skf


def fuse(train_idx):
    action_data = pharma[chemical][action][0].data[train_idx]
    pubmed_data = pharma[chemical][pmid][0].data[train_idx]
    depositor_data = pharma[chemical][depositor][0].data[train_idx]
    fingerprint_data = pharma[chemical][fingerprint][0].data[train_idx]
    depo_cat_data = pharma[depositor][depo_cat][0].data
    chemical_data = pharma[chemical][chemical][0].data[train_idx, :][:, train_idx]
    relations = [
        skf.Relation(action_data, chemical, action),
        skf.Relation(pubmed_data, chemical, pmid),
        skf.Relation(depositor_data, chemical, depositor),
        skf.Relation(fingerprint_data, chemical, fingerprint),
        skf.Relation(depo_cat_data, depositor, depo_cat),
        skf.Relation(chemical_data, chemical, chemical)]
    fusion_graph = skf.FusionGraph(relations)

    fuser = skf.Dfmf(max_iter=200, init_type="random_vcol", random_state=0)
    fuser.fuse(fusion_graph)
    return fuser


def profile(fuser, transformer):
    X = []
    for obj_type in pharma.object_types:
        for c in fuser.chain(chemical, obj_type):
            cf = [fuser.backbone(fuser.fusion_graph[c[i]][c[i+1]][0]) for i in range(len(c)-1)]
            bb = reduce(np.dot, cf) if cf != [] else []
            chem_factor = transformer.factor(chemical)
            chained_profile = np.dot(chem_factor, bb) if bb != [] else chem_factor
            X.append(chained_profile)
    X = np.hstack(X)
    return X


def transform(fuser, test_idx):
    pubmed_data = pharma[chemical][pmid][0].data[test_idx]
    depositor_data = pharma[chemical][depositor][0].data[test_idx]
    fingerprint_data = pharma[chemical][fingerprint][0].data[test_idx]
    chemical_data = pharma[chemical][chemical][0].data[test_idx, :][:, test_idx]
    relations = [
        skf.Relation(pubmed_data, chemical, pmid),
        skf.Relation(depositor_data, chemical, depositor),
        skf.Relation(fingerprint_data, chemical, fingerprint),
        skf.Relation(chemical_data, chemical, chemical)]
    fusion_graph = skf.FusionGraph(relations)

    transformer = skf.DfmfTransform(max_iter=200, init_type="random_vcol", random_state=0)
    transformer.transform(chemical, fusion_graph, fuser)
    return transformer


def predict_action(train_idx, test_idx, action_idx):
    fuser = fuse(train_idx)
    X_train = profile(fuser, fuser)
    y_train = pharma[chemical][action][0].data[train_idx, action_idx]
    clf = ensemble.RandomForestClassifier(n_estimators=200)
    clf.fit(X_train, y_train)

    transformer = transform(fuser, test_idx)
    X_test = profile(fuser, transformer)
    y_pred = clf.predict_proba(X_test)[:, 1]
    return y_pred


def main():
    n_folds = 10
    n_chemicals, n_actions = pharma[chemical][action][0].data.shape
    for t, action_idx in enumerate(range(n_actions)):
        y_true = pharma[chemical][action][0].data[:, action_idx]
        cls_size = int(y_true.sum())
        if cls_size > n_chemicals - 20 or cls_size < 20:
            continue

        cv = cross_validation.StratifiedKFold(y_true, n_folds=n_folds)
        y_pred = np.zeros_like(y_true)
        for i, (train_idx, test_idx) in enumerate(cv):
            y_pred[test_idx] = predict_action(train_idx, test_idx, action_idx)

        action_auc = metrics.roc_auc_score(y_true, y_pred)
        action_name = pharma[chemical][action][0].col_names[action_idx]
        print("(%2d/%2d): %-30s AUC: %0.4f" % (t+1, n_actions, action_name, action_auc))


if __name__ == "__main__":
    pharma = datasets.load_pharma()
    action = pharma.get_object_type('Action')
    pmid = pharma.get_object_type('PMID')
    depositor = pharma.get_object_type('Depositor')
    fingerprint = pharma.get_object_type('Fingerprint')
    depo_cat = pharma.get_object_type('Depositor category')
    chemical = pharma.get_object_type('Chemical')

    main()



"""
==========================================================================
Fusion of three data sources for gene function prediction in Dictyostelium
==========================================================================

Fuse three data sets: gene expression data (Miranda et al., 2013, PLoS One),
slim gene annotations from Gene Ontology and protein-protein interaction
network from STRING database.

Learnt latent matrix factors are utilized for the prediction of slim GO
terms in Dictyostelium genes that are unavailable in the training phase.

This example demonstrates how chaining of fused latent matrices
along the paths of the fusion graph can be used together with an
established prediction model.
"""
print(__doc__)

from functools import reduce

from sklearn import cross_validation, ensemble, metrics
import numpy as np

from skfusion import datasets
from skfusion import fusion as skf


def fuse(train_idx):
    relations = [
        skf.Relation(dicty[gene][go_term][0].data[train_idx, :], gene, go_term),
        skf.Relation(dicty[gene][exp_cond][0].data[train_idx, :], gene, exp_cond),
        skf.Relation(dicty[gene][gene][0].data[train_idx, :][:, train_idx], gene, gene)]
    fusion_graph = skf.FusionGraph(relations)

    fuser = skf.Dfmf(max_iter=50, init_type="random_vcol")
    fuser.fuse(fusion_graph)
    return fuser, fusion_graph


def profile(fuser, transformer):
    X = []
    for obj_type in dicty.object_types:
        for c in fuser.chain(gene, obj_type):
            if obj_type == go_term: continue
            cf = [fuser.backbone(fuser.fusion_graph[c[i]][c[i+1]][0]) for i in range(len(c)-1)]
            bb = reduce(np.dot, cf) if cf != [] else []
            gene_factor = transformer.factor(gene)
            obj_factor = fuser.factor(obj_type)
            chained_profile = np.dot(gene_factor, np.dot(bb, obj_factor.T)) \
                if bb != [] else gene_factor
            X.append(chained_profile)
    X = np.hstack(X)
    return X


def transform(fuser, test_idx):
    relations = [
        skf.Relation(dicty[gene][exp_cond][0].data[test_idx, :], gene, exp_cond),
        skf.Relation(dicty[gene][gene][0].data[test_idx, :][:, test_idx], gene, gene)]
    fusion_graph = skf.FusionGraph(relations)
    transformer = skf.DfmfTransform(max_iter=50, init_type="random_vcol")
    transformer.transform(gene, fusion_graph, fuser)
    return transformer


def predict_term(train_idx, test_idx, term_idx):
    fuser, fuser_graph = fuse(train_idx)
    X_train = profile(fuser, fuser)
    y_train = dicty[gene][go_term][0].data[train_idx, term_idx]
    clf = ensemble.RandomForestClassifier(n_estimators=200)
    clf.fit(X_train, y_train)
    transformer = transform(fuser, test_idx)
    X_test = profile(fuser, transformer)
    y_pred = clf.predict_proba(X_test)[:, 1]
    return y_pred


def main():
    n_folds = 10
    n_genes, n_terms = dicty[gene][go_term][0].data.shape
    for t, term_idx in enumerate(range(n_terms)):
        y_true = dicty[gene][go_term][0].data[:, term_idx]
        cls_size = int(y_true.sum())
        if cls_size > n_genes - 20 or cls_size < 20:
            continue

        cv = cross_validation.StratifiedKFold(y_true, n_folds=n_folds)
        y_pred = np.zeros_like(y_true)
        for i, (train_idx, test_idx) in enumerate(cv):
            y_pred[test_idx] = predict_term(train_idx, test_idx, term_idx)

        term_auc = metrics.roc_auc_score(y_true, y_pred)
        term = dicty[gene][go_term][0].col_names[term_idx]
        print("(%2d/%2d): %10s AUC: %5.4f" % (t+1, n_terms, term, term_auc))


if __name__ == "__main__":
    dicty = datasets.load_dicty()
    gene = dicty.get_object_type("Gene")
    go_term = dicty.get_object_type("GO term")
    exp_cond = dicty.get_object_type("Experimental condition")

    main()



"""
============================================
Data Fusion by Matrix Factorization (`dfmf`)
============================================
"""

import logging
from operator import add
from collections import defaultdict
from functools import reduce

import numpy as np
import scipy.linalg as spla
from joblib import Parallel, delayed

from ._init import initialize


def __bdot(A, B, i, j, obj_types):
    entry = []
    if isinstance(list(A.values())[0], list):
        for l in range(len(A.get((i, j), []))):
            ll = [np.dot(A[i, k][l], B[k, j]) for k in obj_types
                  if (i, k) in A and (k, j) in B]
            if len(ll) > 0:
                tmp = reduce(add, ll)
                entry.append(np.nan_to_num(tmp))
    elif isinstance(list(B.values())[0], list):
        for l in range(len(B.get((i, j), []))):
            ll = [np.dot(A[i, k], B[k, j][l]) for k in obj_types
                  if (i, k) in A and (k, j) in B]
            if len(ll) > 0:
                tmp = reduce(add, ll)
                entry.append(np.nan_to_num(tmp))
    else:
        ll = [np.dot(A[i, k], B[k, j]) for k in obj_types
              if (i, k) in A and (k, j) in B]
        if len(ll) > 0:
            entry = reduce(add, ll)
            entry = np.nan_to_num(entry)
    return i, j, entry


def _par_bdot(A, B, obj_types, verbose, n_jobs):
    """Parallel block matrix multiplication.

    Parameters
    ----------
    A : dictionary of array-like objects
        Block matrix.

    B : dictionary of array-like objects
        Block matrix.

    obj_types : array-like
        Identifiers of object types.

    verbose : int
         The amount of verbosity.

    n_jobs: int (default=1)
        Number of jobs to run in parallel

    Returns
    -------
    C : dictionary of array-like objects
        Matrix product, A*B.
    """
    parallelizer = Parallel(n_jobs=n_jobs, max_nbytes=1e3, verbose=verbose,
                            backend="multiprocessing")
    task_iter = (delayed(__bdot)(A, B, i, j, obj_types)
                 for i in obj_types for j in obj_types)
    entries = parallelizer(task_iter)
    C = {(i, j): entry for i, j, entry in entries if entry != []}
    return C


def _transpose(A):
    """Block matrix transpose.

    Parameters
    ----------
    A : dictionary of array-like objects
        Block matrix.

    Returns
    -------
    At : dictionary of array-like objects
        Block matrix with each of its block transposed.
    """
    At = {k: V.T for k, V in A.items()}
    return At


def count_objects(obj_types, R):
    """Count objects of each object type.

    Parameters
    ----------
    obj_types : set-like
        Object types

    R : dictionary of array-like objects
        Relation matrices

    Returns
    -------
    obj_type2n_obj : dictionary of int
        Number of objects per object type
    """
    obj_type2n_obj = {}
    for r in R:
        i, j = r
        for l in range(len(R[r])):
            for ax, obj_type in enumerate([i, j]):
                ni = obj_type2n_obj.get(obj_type, R[i,j][l].shape[ax])
                if ni != R[i,j][l].shape[ax]:
                    logging.critical("Relation matrix R_(%s,%s) dimension "
                                 "mismatch" % (i, j))
                obj_type2n_obj[obj_type] = ni

    if set(obj_types) != set(obj_type2n_obj.keys()):
        logging.critical("Object type specification mismatch")
    return obj_type2n_obj


def dfmf(R, Theta, obj_types, obj_type2rank,
         max_iter=10, init_type="random_vcol", stopping=None, stopping_system=None,
         verbose=0, compute_err=False, callback=None, random_state=None, n_jobs=1):
    """Data fusion by matrix factorization.

    Parameters
    ----------
    R : dictionary of array-like objects
        Relation matrices.

    Theta : dictionary of array-like objects
        Constraint matrices.

    obj_types : array-like
        Identifiers of object types.

    obj_type2rank : dict-like
        Factorization ranks of object types.

    max_iter : int, optional (default=10)
        Maximum number of iterations to perform.

    init_type : string, optional (default="random_vcol")
        The algorithm to initialize latent matrix factors.

    stopping : tuple (target_matrix, eps), optional (default=None)
        Stopping criterion. Terminate iteration if the reconstruction
        error of target matrix improves by less than eps.

    stopping_system : float, optional (default=None)
        Stopping criterion. Terminate iteration if the reconstruction
        error of the fused system improves by less than eps. compute_err is
        to True to compute the error of the fused system.

    verbose : int, optional (default=0)
         The amount of verbosity. Larger value indicates greater verbosity.

    compute_err : bool, optional (default=False)
        Compute the reconstruction error of every relation matrix if True.

    callback : callable, optional
        An optional user-supplied function to call after each iteration. Called
        as callback(G, S, cur_iter), where S and G are current latent estimates.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance
        used by np.random.

    n_jobs: int (default=1)
        Number of jobs to run in parallel

    Returns
    -------
    G :
    S :
    """
    verbose1 = verbose
    verbose = 50 - verbose
    logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s",
                        datefmt="%m/%d/%Y %I:%M:%S %p", level=verbose)

    obj_type2n_obj = count_objects(obj_types, R)
    R_to_init = {k: V[0] for k, V in R.items()}
    G = initialize(obj_types, obj_type2n_obj, obj_type2rank, R_to_init,
                   init_type, random_state)
    S = None

    if stopping:
        err = (None, None)
    if stopping_system:
        err_system = (None, None)
        compute_err = True

    logging.info("Solving for Theta_p and Theta_n")
    Theta_p, Theta_n = defaultdict(list), defaultdict(list)
    for r, thetas in Theta.items():
        for theta in thetas:
            t = theta > 0
            Theta_p[r].append(np.multiply(t, theta))
            Theta_n[r].append(np.multiply(t-1, theta))

    obj = []

    for iter in range(max_iter):
        if iter > 1 and stopping and err[1]-err[0] < stopping[1]:
            logging.info("Early stopping: target matrix change < %5.4f" \
                      % stopping[1])
            break
        if iter > 1 and stopping_system and \
                                err_system[1]-err_system[0] < stopping_system:
            logging.info("Early stopping: matrix system change < %5.4f" \
                      % stopping_system)
            break

        logging.info("Factorization iteration: %d" % iter)

        #######################################################################
        ########################### General Update ############################

        pGtG = {}
        for r in G:
            logging.info("Computing GrtGr: %s" % str(r))
            GrtGr = np.nan_to_num(np.dot(G[r].T, G[r]))
            pGtG[r] = spla.pinv(GrtGr)

        logging.info("Start to update S")

        tmp1 = _par_bdot(G, pGtG, obj_types, verbose1, n_jobs)
        tmp2 = _par_bdot(R, tmp1, obj_types, verbose1, n_jobs)
        tmp3 = _par_bdot(_transpose(G), tmp2, obj_types, verbose1, n_jobs)
        S = _par_bdot(pGtG, tmp3, obj_types, verbose1, n_jobs)

        #######################################################################
        ########################### General Update ############################

        logging.info("Start to update G")

        G_enum = {r: np.zeros(Gr.shape) for r, Gr in G.items()}
        G_denom = {r: np.zeros(Gr.shape) for r, Gr in G.items()}

        for r in R:
            i, j = r
            for l in range(len(R[r])):
                logging.info("Update G due to R_%s,%s^(%d)" % (i, j, l))

                tmp1 = np.dot(R[i, j][l], np.dot(G[j, j], S[i, j][l].T))
                tmp1 = np.nan_to_num(tmp1)
                t = tmp1 > 0
                tmp1p = np.multiply(t, tmp1)
                tmp1n = np.multiply(t-1, tmp1)

                tmp2 = np.dot(S[i, j][l], np.dot(G[j, j].T, np.dot(G[j, j], S[i, j][l].T)))
                tmp2 = np.nan_to_num(tmp2)
                t = tmp2 > 0
                tmp2p = np.multiply(t, tmp2)
                tmp2n = np.multiply(t-1, tmp2)

                tmp4 = np.dot(R[i, j][l].T, np.dot(G[i, i], S[i, j][l]))
                tmp4 = np.nan_to_num(tmp4)
                t = tmp4 > 0
                tmp4p = np.multiply(t, tmp4)
                tmp4n = np.multiply(t-1, tmp4)

                tmp5 = np.dot(S[i, j][l].T, np.dot(G[i, i].T, np.dot(G[i, i], S[i, j][l])))
                tmp5 = np.nan_to_num(tmp5)
                t = tmp5 > 0
                tmp5p = np.multiply(t, tmp5)
                tmp5n = np.multiply(t-1, tmp5)

                G_enum[i, i] += tmp1p + np.dot(G[i, i], tmp2n)
                G_denom[i, i] += tmp1n + np.dot(G[i, i], tmp2p)

                G_enum[j, j] += tmp4p + np.dot(G[j, j], tmp5n)
                G_denom[j, j] += tmp4n + np.dot(G[j, j], tmp5p)

        logging.info("Update of G due to constraint matrices")
        for r, thetas_p in Theta_p.items():
            logging.info("Considering Theta pos. %s" % str(r))
            for theta_p in thetas_p:
                G_denom[r] += np.dot(theta_p, G[r])
        for r, thetas_n in Theta_n.items():
            logging.info("Considering Theta neg. %s" % str(r))
            for theta_n in thetas_n:
                G_enum[r] += np.dot(theta_n, G[r])

        for r in G:
            G[r] = np.multiply(G[r], np.sqrt(
                np.divide(G_enum[r], np.maximum(G_denom[r], np.finfo(np.float).eps))))

        #######################################################################
        ######################## Reconstruction Error #########################

        if stopping:
            target, eps = stopping
            err = (np.linalg.norm(R[target]-np.dot(G[target[0],target[0]],
                    np.dot(S[target], G[target[1],target[1]].T))), err[0])

        if compute_err:
            s = 0
            for r in R:
                i, j = r
                for l in range(len(R[r])):
                    Rij_app = np.dot(G[i, i], np.dot(S[i, j][l], G[j, j].T))
                    r_err = np.linalg.norm(R[r][l]-Rij_app, "fro")
                    logging.info("Relation R_%s,%s^(%d) norm difference: " \
                              "%5.4f" % (i, j, l, r_err))
                    s += r_err
            logging.info("Error (objective function value): %5.4f" % s)
            obj.append(s)
            if stopping_system:
                err_system = (s, err_system[0])

        if callback:
            callback(G, S, iter)

    if compute_err:
        logging.info("Violations of optimization objective: %d/%d " % (
            int(np.sum(np.diff(obj) > 0)), len(obj)))
    return G, S


def transform(R_ij, Theta_i, target_obj_type, obj_type2rank, G, S,
              max_iter=10, init_type="random_c",
              stopping=None, stopping_system=None, verbose=0,
              compute_err=False, callback=None, random_state=None):
    verbose = 50 - verbose
    logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s",
                        datefmt="%m/%d/%Y %I:%M:%S %p", level=verbose)
    if isinstance(random_state, np.random.RandomState):
        random_state = random_state
    else:
        random_state = np.random.RandomState(random_state)

    n_targets = [R_ij[i,j][0].shape[0 if target_obj_type == i else 1] for i, j in R_ij]
    if len(set(n_targets)) > 1:
        logging.critical("Target object type: %s size mismatch" % target_obj_type)
    n_targets = n_targets[0]
    R_to_init = {k: V[0] for k, V in R_ij.items()}
    Gx = initialize(
        [target_obj_type], {target_obj_type: n_targets}, obj_type2rank,
        R_to_init, init_type, random_state)
    G_i = Gx[target_obj_type, target_obj_type]

    if stopping:
        err = (None, None)
    if stopping_system:
        err_system = (None, None)
        compute_err = True

    Theta_p, Theta_n = [], []
    for r, thetas in Theta_i.items():
        for theta in thetas:
            t = theta > 0
            Theta_p.append(np.multiply(t, theta))
            Theta_n.append(np.multiply(t-1, theta))

    obj = []

    for iter in range(max_iter):
        if iter > 1 and stopping and abs(err[1]-err[0]) < stopping[1]:
            logging.info("Early stopping: target matrix change < %5.4f" \
                      % stopping[1])
            break
        if iter > 1 and stopping_system and \
                                err_system[1]-err_system[0] < stopping_system:
            logging.info("Early stopping: matrix system change < %5.4f" \
                      % stopping_system)
            break

        logging.info("Factorization iteration: %d" % iter)

        #######################################################################
        ########################### General Update ############################

        logging.info("Start to update G")

        G_enum = np.zeros(G_i.shape)
        G_denom = np.zeros(G_i.shape)

        for r in R_ij:
            i, j = r
            for l in range(len(R_ij[r])):
                logging.info("Update G due to R_%s,%s^(%d)" % (i, j, l))
                if i is target_obj_type:

                    tmp1 = np.dot(R_ij[i, j][l], np.dot(G[j, j], S[i, j][l].T))
                    t = tmp1 > 0
                    tmp1p = np.multiply(t, tmp1)
                    tmp1n = np.multiply(t-1, tmp1)

                    tmp2 = np.dot(S[i, j][l], np.dot(G[j, j].T, np.dot(G[j, j], S[i, j][l].T)))
                    t = tmp2 > 0
                    tmp2p = np.multiply(t, tmp2)
                    tmp2n = np.multiply(t-1, tmp2)

                    G_enum += tmp1p + np.dot(G_i, tmp2n)
                    G_denom += tmp1n + np.dot(G_i, tmp2p)

                if j is target_obj_type:
                    tmp4 = np.dot(R_ij[i, j][l].T, np.dot(G[i, i], S[i, j][l]))
                    t = tmp4 > 0
                    tmp4p = np.multiply(t, tmp4)
                    tmp4n = np.multiply(t-1, tmp4)

                    tmp5 = np.dot(S[i, j][l].T, np.dot(G[i, i].T, np.dot(G[i, i], S[i, j][l])))
                    t = tmp5 > 0
                    tmp5p = np.multiply(t, tmp5)
                    tmp5n = np.multiply(t-1, tmp5)

                    G_enum += tmp4p + np.dot(G_i, tmp5n)
                    G_denom += tmp4n + np.dot(G_i, tmp5p)

        logging.info("Update of G due to constraint matrices")
        for theta_p in Theta_p:
            G_denom += np.dot(theta_p, G_i)
        for theta_n in Theta_n:
            G_enum += np.dot(theta_n, G_i)

        G_i = np.multiply(G_i, np.sqrt(
            np.divide(G_enum, np.maximum(G_denom, np.finfo(np.float).eps))))

        #######################################################################
        ######################## Reconstruction Error #########################

        if compute_err:
            s = 0
            for r in R_ij:
                i, j = r
                for l in range(len(R_ij[r])):
                    if i is target_obj_type:
                        Rij_app = np.dot(G_i, np.dot(S[i, j][l], G[j, j].T))
                        r_err = np.linalg.norm(R_ij[r][l]-Rij_app, "fro")
                    if j is target_obj_type:
                        Rij_app = np.dot(G[i, i], np.dot(S[i, j][l], G_i.T))
                        r_err = np.linalg.norm(R_ij[r][l]-Rij_app, "fro")
                    logging.info("Relation R_%s,%s^(%d) norm difference: " \
                          "%5.4f" % (i, j, l, r_err))
                    s += r_err
            logging.info("Error (objective function value): %5.4f" % s)
            obj.append(s)
            if stopping_system:
                err_system = (s, err_system[0])

        if callback:
            callback(G_i, iter)

    if compute_err:
        logging.info("Violations of optimization objective: %d/%d " % (
            int(np.sum(np.diff(obj) > 0)), len(obj)))
    return G_i


if 0:
    # toy example for factorizing new examples
    obj_types = [0, 1, 2]
    n1, n2, n3 = 400, 80, 264
    c1, c2, c3 = 30, 40, 50
    obj_type2rank = {0: c1, 1: c2, 2: c3}

    rnds = np.random.RandomState(0)

    R12 = 5 * rnds.rand(n1, n2)
    R13 = 20 * rnds.rand(n1, n3)
    R21 = R12.T
    R23 = 20 * rnds.rand(n2, n3)

    Theta = {}
    R = {(0, 1): [R12], (0, 2): [R13], (1, 2): [R23], (1, 0): [R21]}
    print("Inference")
    G, S = dfmf(
        R, Theta, obj_types, obj_type2rank, verbose=0, max_iter=10,
        init_type="random", compute_err=True, random_state=rnds)

    # new data
    o_n2_z3 = 20 * rnds.rand(3, n3)
    o_n2_z1 = 5 * rnds.rand(3, n1)

    print("We got some new data!")
    x_G2 = transform(
        {(1,0): [o_n2_z1], (0,1): [o_n2_z1.T], (1,2): [o_n2_z3]}, {},
        1, obj_type2rank, G, S, max_iter=10, init_type="random",
        compute_err=True, stopping=None, random_state=rnds)

    R12_hat = np.dot(G[0,0], np.dot(S[0,1][0], G[1,1].T))
    R21_hat = np.dot(G[1,1], np.dot(S[1,0][0], G[0,0].T))
    R23_hat = np.dot(G[1,1], np.dot(S[1,2][0], G[2,2].T))

    x_R12_hat = np.dot(G[0,0], np.dot(S[0,1][0], x_G2.T))
    x_R21_hat = np.dot(x_G2, np.dot(S[1,0][0], G[0,0].T))
    x_R23_hat = np.dot(x_G2, np.dot(S[1,2][0], G[2,2].T))

    err = np.linalg.norm(R23[0, :] - R23_hat[0, :])
    print("Error in R23 of object that existed before: %5.3f" % err)
    x_err = np.linalg.norm((o_n2_z3 - x_R23_hat)[0, :])
    print("Error in R23 of newly added object: %5.3f" % x_err)

    err = np.linalg.norm(R12[:, -1] - R12_hat[:, -1])
    print("Error in R12 of object that existed before: %5.3f" % err)
    x_err = np.linalg.norm((x_R12_hat - o_n2_z1.T)[:, 0])
    print("Error in R12 of newly added object: %5.3f" % x_err)

    err = np.linalg.norm(R21[0, :] - R21_hat[0, :])
    print("Error in R21 of object that existed before: %5.3f" % err)
    x_err = np.linalg.norm((x_R21_hat - o_n2_z1)[0, :])
    print("Error in R21 of newly added object: %5.3f" % x_err)



import unittest

import numpy as np
from skfusion.fusion import Relation, ObjectType, \
    FusionGraph, Dfmf, DfmfTransform, Dfmc


class TestMultipleRelations(unittest.TestCase):
    def test_dfmf(self):
        rnds = np.random.RandomState(0)
        R12_1 = np.random.rand(30, 30)
        R12_2 = np.random.rand(30, 30)
        R13 = np.random.rand(30, 20)

        t1 = ObjectType('type1', 30)
        t2 = ObjectType('type2', 30)
        t3 = ObjectType('type3', 20)
        relations = [Relation(R12_1, t1, t2), Relation(R12_2, t1, t2),
                     Relation(R13, t1, t3)]
        fusion_graph = FusionGraph()
        fusion_graph.add_relations_from(relations)
        self.assertEqual(len(fusion_graph.relations), 3)
        self.assertEqual(len(fusion_graph.object_types), 3)

        fuser = Dfmf(init_type='random', random_state=rnds).fuse(fusion_graph)
        self.assertEqual(fuser.backbone(relations[0]).shape, (30, 30))
        self.assertEqual(fuser.backbone(relations[1]).shape, (30, 30))
        self.assertEqual(fuser.backbone(relations[2]).shape, (30, 20))
        G1 = fuser.factor(t1)
        G2 = fuser.factor(t2)
        S12_1 = fuser.backbone(relations[0])
        S12_2 = fuser.backbone(relations[1])
        R12_1_hat = np.dot(G1, np.dot(S12_1, G2.T))
        R12_2_hat = np.dot(G1, np.dot(S12_2, G2.T))
        np.testing.assert_almost_equal(fuser.complete(relations[0]), R12_1_hat)
        np.testing.assert_almost_equal(fuser.complete(relations[1]), R12_2_hat)

    def test_dfmc(self):
        rnds = np.random.RandomState(0)
        R12_1 = np.random.rand(30, 30)
        R12_2 = np.random.rand(30, 30)
        R13 = np.random.rand(30, 20)

        t1 = ObjectType('type1', 30)
        t2 = ObjectType('type2', 30)
        t3 = ObjectType('type3', 20)
        relations = [Relation(R12_1, t1, t2), Relation(R12_2, t1, t2),
                     Relation(R13, t1, t3)]
        fusion_graph = FusionGraph()
        fusion_graph.add_relations_from(relations)
        self.assertEqual(len(fusion_graph.relations), 3)
        self.assertEqual(len(fusion_graph.object_types), 3)

        fuser = Dfmc(init_type='random', random_state=rnds).fuse(fusion_graph)
        self.assertEqual(fuser.backbone(relations[0]).shape, (30, 30))
        self.assertEqual(fuser.backbone(relations[1]).shape, (30, 30))
        self.assertEqual(fuser.backbone(relations[2]).shape, (30, 20))
        G1 = fuser.factor(t1)
        G2 = fuser.factor(t2)
        S12_1 = fuser.backbone(relations[0])
        S12_2 = fuser.backbone(relations[1])
        R12_1_hat = np.dot(G1, np.dot(S12_1, G2.T))
        R12_2_hat = np.dot(G1, np.dot(S12_2, G2.T))
        np.testing.assert_almost_equal(fuser.complete(relations[0]), R12_1_hat)
        np.testing.assert_almost_equal(fuser.complete(relations[1]), R12_2_hat)


if __name__ == "__main__":
    unittest.main()



import os
import unittest


def suite():
    test_dir = os.path.dirname(__file__)
    return unittest.TestLoader().discover(test_dir)

test_suite = suite()

if __name__ == '__main__':
    unittest.main(defaultTest='suite')



import unittest

import numpy as np
from skfusion.fusion import Relation, ObjectType, \
    FusionGraph


class TestFusionGraph(unittest.TestCase):
    def setUp(self):
        rnds = np.random.RandomState(0)
        X = rnds.rand(30, 30)
        self.t1 = ObjectType('Type 1', 10)
        self.t2 = ObjectType('Type 2', 10)
        self.t3 = ObjectType('Type 3', 10)
        self.t4 = ObjectType('Type 4', 10)
        self.t5 = ObjectType('Type 5', 10)
        self.relations1 = [
            Relation(X, self.t1, self.t2, name='Test1'), Relation(X, self.t2, self.t3),
            Relation(X, self.t3, self.t4), Relation(X, self.t4, self.t5),
            Relation(X, self.t3, self.t5), Relation(X, self.t5, self.t1)]

        self.relations2 = [
            Relation(X, self.t1, self.t2, name='Test2'), Relation(X, self.t1, self.t2),
            Relation(X, self.t2, self.t3), Relation(X, self.t3, self.t4),
            Relation(X, self.t4, self.t5), Relation(X, self.t3, self.t5),
            Relation(X, self.t5, self.t1), Relation(X, self.t4, self.t4),
            Relation(X, self.t4, self.t4, name='Test3'), Relation(X, self.t5, self.t5)]

    def test_drawing(self):
        fusion_graph = FusionGraph()
        fusion_graph.add_relations_from(self.relations1)
        # fusion_graph.draw_graphviz('test_fusion_graph.pdf')

    def test_manipulation(self):
        fusion_graph = FusionGraph()
        fusion_graph.add_relations_from(self.relations2)

        self.assertEqual(fusion_graph['Test2'], self.relations2[0])
        self.assertEqual(fusion_graph['Test3'], self.relations2[8])

        self.assertEqual(fusion_graph.n_object_types, 5)
        self.assertEqual(fusion_graph.n_relations, 10)

        fusion_graph.remove_relation(self.relations2[6])
        self.assertEqual(fusion_graph.n_object_types, 5)
        self.assertEqual(fusion_graph.n_relations, 9)

        fusion_graph.remove_relations_from([
            self.relations2[9], self.relations2[4], self.relations2[5]])
        self.assertEqual(fusion_graph.n_object_types, 4)
        self.assertEqual(fusion_graph.n_relations, 6)

    def test_inspection(self):
        fusion_graph = FusionGraph(self.relations2)
        self.assertEqual(set(fusion_graph.in_relations(self.t1)), {self.relations2[6]})
        self.assertEqual(set(fusion_graph.out_relations(self.t1)), set(self.relations2[:2]))
        out_nbs = {self.relations2[4], self.relations2[7], self.relations2[8]}
        self.assertEqual(set(fusion_graph.out_relations(self.t4)), out_nbs)

    def test_retrieval(self):
        fusion_graph = FusionGraph(self.relations2)
        self.assertEqual(fusion_graph.get_object_type('Type 1'), self.t1)
        self.assertEqual(list(fusion_graph.get_relations(self.t1, self.t2)), self.relations2[:2])
        self.assertEqual(fusion_graph[self.t1][self.t2], self.relations2[:2])
        out_degree1 = len(list(fusion_graph.out_relations(self.t4)))
        out_degree2 = sum(len(rels) for rels in fusion_graph[self.t4].values())
        self.assertEqual(out_degree1, out_degree2)

    def test_removal_single_relation(self):
        fusion_graph = FusionGraph()
        fusion_graph.add_relation(self.relations1[0])
        self.assertEqual(fusion_graph.n_relations, 1)
        self.assertEqual(fusion_graph.n_object_types, 2)
        fusion_graph.remove_relation(self.relations1[0])
        self.assertEqual(fusion_graph.n_relations, 0)
        self.assertEqual(fusion_graph.n_object_types, 0)

    def test_removal_of_loops(self):
        fusion_graph = FusionGraph()
        fusion_graph.add_relation(self.relations2[-1])
        self.assertEqual(fusion_graph.n_relations, 1)
        self.assertEqual(fusion_graph.n_object_types, 1)
        fusion_graph.remove_relation(self.relations2[-1])
        self.assertEqual(fusion_graph.n_relations, 0)
        self.assertEqual(fusion_graph.n_object_types, 0)

    def test_get_names_by_object_type(self):
        rnds = np.random.RandomState(0)
        X = rnds.rand(10, 10)
        t1_names = list('ABCDEFGHIJ')
        t2_names = list('KLMNOPQRST')

        rel = Relation(X, name='Test',
                       row_type=self.t1, row_names=t1_names,
                       col_type=self.t2, col_names=t2_names)
        rel2 = Relation(X, name='Test2',
                        row_type=self.t2, row_names=t2_names,
                        col_type=self.t3)
        fusion_graph = FusionGraph()
        fusion_graph.add_relation(rel)
        fusion_graph.add_relation(rel2)

        self.assertEqual(fusion_graph.get_names(self.t1), t1_names)
        self.assertEqual(fusion_graph.get_names(self.t2), t2_names)
        t3_names = fusion_graph.get_names(self.t3)
        self.assertEqual(len(t3_names), 10)

    def test_get_object_type_metadata(self):
        rnds = np.random.RandomState(0)
        X = rnds.rand(10, 10)
        a, b, c = list('ABCDEFGHIJ'), list('0123456789'), list('KLMNOPQRST')
        t1_metadata = [{'a': x} for x in a]
        t2_metadata = [{'b': x} for x in b]
        t2_metadata2 = [{'d': x} for x in b]

        rel = Relation(X, name='Test',
                       row_type=self.t1, row_metadata=t1_metadata,
                       col_type=self.t2, col_metadata=t2_metadata)
        rel2 = Relation(X, name='Test2',
                        row_type=self.t2, row_metadata=t2_metadata2,
                        col_type=self.t3)
        fusion_graph = FusionGraph()
        fusion_graph.add_relation(rel)
        fusion_graph.add_relation(rel2)

        def merge(d1, d2):
            d = {}
            d.update(d1)
            d.update(d2)
            return d

        self.assertEqual(fusion_graph.get_metadata(self.t1), t1_metadata)
        self.assertEqual(fusion_graph.get_metadata(self.t2), list(map(merge, t2_metadata, t2_metadata2)))
        t3_metadata = fusion_graph.get_metadata(self.t3)
        self.assertEqual(len(t3_metadata), 10)
        for md in t3_metadata:
            self.assertFalse(md)


if __name__ == "__main__":
    unittest.main()



"""
Base code for handling data sets.
"""
import gzip
import csv
from collections import defaultdict
from os.path import dirname
from os.path import join

import numpy as np

from skfusion.fusion import ObjectType, Relation, FusionGraph


__all__ = ['load_dicty', 'load_pharma', 'load_movielens']


def load_source(source_path, delimiter=',', filling_value='0'):
    """Load and return a data source.

    Parameters
    ----------
    delimiter : str, optional (default=',')
        The string used to separate values. By default, comma acts as delimiter.

    filling_value : variable, optional (default='0')
        The value to be used as default when the data are missing.

    Returns
    -------
    data : DataSource
        Dictionary-like object, the interesting attributes are:
        'data', the data to learn, 'obj1_names', the meaning of row objects,
        'obj2_names', the meaning of column objects.
    """
    module_path = dirname(__file__)
    data_file = gzip.open(join(module_path, 'data', source_path))
    row_names = np.array(next(data_file).decode('utf-8').strip().split(delimiter))
    col_names = np.array(next(data_file).decode('utf-8').strip().split(delimiter))
    data = np.genfromtxt(data_file, delimiter=delimiter, missing_values=[''],
                         filling_values=filling_value)
    return data, row_names, col_names


def load_dicty():
    """Construct fusion graph from molecular biology of Dictyostelium."""
    gene = ObjectType('Gene', 50)
    go_term = ObjectType('GO term', 15)
    exprc = ObjectType('Experimental condition', 5)

    data, rn, cn = load_source(join('dicty', 'dicty.gene_annnotations.csv.gz'))
    ann = Relation(data=data, row_type=gene, col_type=go_term, name='ann',
                   row_names=rn, col_names=cn)
    data, rn, cn = load_source(join('dicty', 'dicty.gene_expression.csv.gz'))
    expr = Relation(data=data, row_type=gene, col_type=exprc, name='expr',
                    row_names=rn, col_names=cn)
    expr.data = np.log(np.maximum(expr.data, np.finfo(np.float).eps))
    data, rn, cn = load_source(join('dicty', 'dicty.ppi.csv.gz'))
    ppi = Relation(data=data, row_type=gene, col_type=gene, name='ppi',
                   row_names=rn, col_names=cn)
    return FusionGraph([ann, expr, ppi])


def load_pharma():
    """Construct fusion graph from the pharmacology domain."""
    action=ObjectType('Action', 5)
    pmid=ObjectType('PMID', 5)
    depositor=ObjectType('Depositor', 5)
    fingerprint=ObjectType('Fingerprint', 20)
    depo_cat=ObjectType('Depositor category', 5)
    chemical=ObjectType('Chemical', 10)

    data, rn, cn = load_source(join('pharma', 'pharma.actions.csv.gz'))
    actions = Relation(data=data, row_type=chemical, col_type=action,
                       row_names=rn, col_names=cn)
    data, rn, cn = load_source(join('pharma', 'pharma.pubmed.csv.gz'))
    pubmed = Relation(data=data, row_type=chemical, col_type=pmid,
                      row_names=rn, col_names=cn)
    data, rn, cn = load_source(join('pharma', 'pharma.depositors.csv.gz'))
    depositors = Relation(data=data, row_type=chemical, col_type=depositor,
                          row_names=rn, col_names=cn)
    data, rn, cn = load_source(join('pharma', 'pharma.fingerprints.csv.gz'))
    fingerprints = Relation(data=data, row_type=chemical, col_type=fingerprint,
                            row_names=rn, col_names=cn)
    data, rn, cn = load_source(join('pharma', 'pharma.depo_cats.csv.gz'))
    depo_cats = Relation(data=data, row_type=depositor, col_type=depo_cat,
                         row_names=rn, col_names=cn)
    data, rn, cn = load_source(join('pharma', 'pharma.tanimoto.csv.gz'))
    tanimoto = Relation(data=data, row_type=chemical, col_type=chemical,
                        row_names=rn, col_names=cn)
    return FusionGraph([actions, pubmed, depositors, fingerprints, depo_cats, tanimoto])


def load_movielens(ratings=True, movie_genres=True, movie_actors=True):
    module_path = join(dirname(__file__), 'data', 'movielens')
    if ratings:
        ratings_data = defaultdict(dict)
        with gzip.open(join(module_path, 'ratings.csv.gz'), 'rt', encoding='utf-8') as f:
            f.readline()
            for line in f:
                line = line.strip().split(',')
                ratings_data[int(line[0])][int(line[1])] = float(line[2])
    else:
        ratings_data = None

    if movie_genres:
        movie_genres_data = {}
        with gzip.open(join(module_path, 'movies.csv.gz'), 'rt', encoding='utf-8') as f:
            f.readline()
            lines = csv.reader(f)
            for line in lines:
                movie_genres_data[int(line[0])] = line[2].split('|')
    else:
        movie_genres_data = None

    if movie_actors:
        movie_actors_data = {}
        with gzip.open(join(module_path, 'actors.csv.gz'), 'rt', encoding='utf-8') as f:
            f.readline()
            lines = csv.reader(f)
            for line in lines:
                movie_actors_data[int(line[0])] = line[2].split('|')
    else:
        movie_actors_data = None
    return ratings_data, movie_genres_data, movie_actors_data



import unittest

import numpy as np
from skfusion.fusion import Relation, ObjectType, \
    FusionGraph, Dfmf, DfmfTransform


class TestDfmf(unittest.TestCase):
    def test_dfmf(self):
        rnds = np.random.RandomState(0)
        R12 = rnds.rand(50, 30)

        t1 = ObjectType('type1', 50)
        t2 = ObjectType('type2', 30)
        relation = Relation(R12, t1, t2)
        fusion_graph = FusionGraph()
        fusion_graph.add_relation(relation)

        fuser = Dfmf(init_type='random', random_state=rnds).fuse(fusion_graph)
        self.assertEqual(fuser.backbone(relation).shape, (50, 30))
        self.assertEqual(fuser.factor(t1).shape, (50, 50))
        self.assertEqual(fuser.factor(t2).shape, (30, 30))
        np.testing.assert_almost_equal(fuser.complete(relation), relation.data)

    def test_infinite(self):
        rnds = np.random.RandomState(0)
        R12 = rnds.rand(50, 30)
        R13 = rnds.rand(50, 10)
        R12 = np.ma.masked_greater(R12, 0.7)
        R12[R12 < 0.1] = np.nan
        R13[R13 < 0.5] = np.inf

        t1 = ObjectType('type1', 50)
        t2 = ObjectType('type2', 30)
        t3 = ObjectType('type3', 10)
        relations = [Relation(R12, t1, t2), Relation(R13, t1, t3)]
        fusion_graph = FusionGraph(relations)

        fuser = Dfmf(init_type='random', random_state=rnds, fill_value=0).fuse(fusion_graph)
        self.assertEqual(fuser.backbone(relations[0]).shape, (50, 30))
        self.assertEqual(fuser.backbone(relations[1]).shape, (50, 10))
        self.assertEqual(fuser.factor(t1).shape, (50, 50))
        self.assertEqual(fuser.factor(t2).shape, (30, 30))
        np.testing.assert_equal(np.sum(np.isfinite(fuser.complete(relations[0]))), R12.size)

    def test_transformation(self):
        R12 = np.random.rand(5, 3)

        t1 = ObjectType('type1', 2)
        t2 = ObjectType('type2', 2)
        relation = Relation(R12, t1, t2)
        fusion_graph = FusionGraph()
        fusion_graph.add_relation(relation)

        rnds = np.random.RandomState(0)
        fuser = Dfmf(init_type='random', random_state=rnds, max_iter=100
                     ).fuse(fusion_graph)

        new_R12 = R12[:2].copy()
        new_graph = FusionGraph([Relation(new_R12, t1, t2)])

        new_rnds = np.random.RandomState(0)
        transformer = DfmfTransform(random_state=new_rnds).transform(
            t1, new_graph, fuser)

        new_G1 = transformer.factor(t1)
        G1 = fuser.factor(t1)
        G2 = fuser.factor(t2)
        S12 = fuser.backbone(relation)
        new_R12_hat = np.dot(new_G1, np.dot(S12, G2.T))
        R12_hat = np.dot(G1, np.dot(S12, G2.T))

        diff_G1 = new_G1 - G1[:2]
        diff_hat = new_R12_hat - R12_hat[:2]
        self.assertLess(np.sum(diff_G1 ** 2) / diff_G1.size, 1e-5)
        self.assertLess(np.sum(diff_hat ** 2) / diff_hat.size, 1e-5)

    def test_preprocessors(self):
        rnds = np.random.RandomState(0)
        R12 = rnds.rand(50, 30)

        t1 = ObjectType('type1', 50)
        t2 = ObjectType('type2', 30)

        def preprocessor(data):
            return np.ones_like(data)

        relation = Relation(R12, t1, t2, preprocessor=preprocessor)
        fusion_graph = FusionGraph()
        fusion_graph.add_relation(relation)

        fuser = Dfmf(init_type='random', random_state=rnds).fuse(fusion_graph)
        self.assertEqual(fuser.backbone(relation).shape, (50, 30))
        self.assertEqual(fuser.factor(t1).shape, (50, 50))
        self.assertEqual(fuser.factor(t2).shape, (30, 30))
        trnf = np.ones_like(relation.data)
        np.testing.assert_almost_equal(fuser.complete(relation), trnf)

    def test_postprocessors(self):
        rnds = np.random.RandomState(0)
        R12 = rnds.rand(50, 30)

        t1 = ObjectType('type1', 50)
        t2 = ObjectType('type2', 30)

        def postprocessor(data):
            return data - np.mean(data)

        relation = Relation(R12, t1, t2, postprocessor=postprocessor)
        fusion_graph = FusionGraph()
        fusion_graph.add_relation(relation)

        fuser = Dfmf(init_type='random', random_state=rnds).fuse(fusion_graph)
        self.assertEqual(fuser.backbone(relation).shape, (50, 30))
        self.assertEqual(fuser.factor(t1).shape, (50, 50))
        self.assertEqual(fuser.factor(t2).shape, (30, 30))
        trnf = relation.data - np.mean(relation.data)
        np.testing.assert_almost_equal(fuser.complete(relation), trnf)


if __name__ == "__main__":
    unittest.main()



"""
=====================================================================
Fusion of user ratings data, movie genres and data about movie actors
=====================================================================

This example demonstrates matrix completion on user-movie preference
data.

DISCLAIMER: Results can be improved if ideas from collaborative
filtering (regularization, user bias, item bias, etc.) are applied.
"""
print(__doc__)

import numpy as np

from skfusion import datasets
from skfusion import fusion as skf


def load_data():
    ratings_data, movies_data, actors_data = datasets.load_movielens()
    n_movies = 1000
    n_actors = 1000

    movies = set(m for val in ratings_data.values() for m in val)
    movies = list(movies)[:n_movies]
    movie2id = {m: i for i, m, in enumerate(movies)}
    user2id = {u: i for i, u in enumerate(ratings_data.keys())}
    genres = set(e for val1 in movies_data.values() for e in val1)
    genre2id = {g: i for i, g in enumerate(genres)}
    actors = set(e for m, val1 in actors_data.items()
                 for e in val1 if m in movie2id)
    actors = list(actors)[:n_actors]
    actor2id = {a: i for i, a in enumerate(actors)}

    R12_true = -1 * np.ones((len(user2id), len(movie2id)))
    print('User ratings: {}'.format(str(R12_true.shape)))
    for user in ratings_data:
        for movie in ratings_data[user]:
            if movie not in movie2id: continue
            R12_true[user2id[user], movie2id[movie]] = ratings_data[user][movie]
    R12_true = np.ma.masked_equal(R12_true, -1)
    R12_true = scale(R12_true, 0, 1)

    R23 = np.zeros((len(movie2id), len(genre2id)))
    print('Movie genres: {}'.format(str(R23.shape)))
    for movie in movies_data:
        for genre in movies_data[movie]:
            if movie not in movie2id: continue
            R23[movie2id[movie], genre2id[genre]] = 1.

    R24 = np.zeros((len(movie2id), len(actor2id)))
    print('Actors played: {}'.format(str(R24.shape)))
    for movie in actors_data:
        for actor in actors_data[movie]:
            if actor not in actor2id: continue
            if movie not in movie2id: continue
            R24[movie2id[movie], actor2id[actor]] = 1.

    frac = 0.9
    R12 = R12_true.copy()
    hide = np.logical_and(np.random.random(R12_true.shape) > frac, ~R12_true.mask)
    R12 = np.ma.masked_where(hide, R12)

    mean_user = np.mean(R12, 1)
    mean_movie = np.mean(R12, 0)
    mean_rating = np.mean(R12)
    means = mean_user, mean_movie, mean_rating

    p = 0.05
    t1 = skf.ObjectType('User', max(int(p*R12.shape[0]), 5))
    t2 = skf.ObjectType('Movie', max(int(p*R12.shape[1]), 5))
    t3 = skf.ObjectType('Genre', max(int(p*R23.shape[1]), 5))
    t4 = skf.ObjectType('Actor', max(int(p*R24.shape[1]), 5))

    # L2 regularization can be added via diagonal constraint matrices
    relations = [skf.Relation(R12, t1, t2, name='User ratings'),
                 skf.Relation(R23, t2, t3, name='Movie genres'),
                 skf.Relation(R24, t2, t4, name='Movie actors')]
    graph = skf.FusionGraph(relations)
    print('Ranks:', ''.join(['\n{}: {}'.format(o.name, o.rank)
                             for o in graph.object_types]))

    graph_small = skf.FusionGraph([skf.Relation(R12, t1, t2, name='User ratings')])

    return R12_true, hide, means, graph, graph_small


def rmse(y_true, y_pred):
    return np.sqrt(np.sum((y_true - y_pred)**2) / y_true.size)


def scale(X, amin, amax):
    return (X - X.min()) / (X.max() - X.min()) * (amax - amin) + amin


def main():
    R12_true, hidden, means, graph, graph_small = load_data()
    mean_user, mean_movie, mean_rating = means
    n_users, n_movies = R12_true.shape

    # mean rating
    score = rmse(R12_true[hidden], mean_rating)
    print('RMSE(mean rating): {}'.format(score))

    # mean user
    R12_pred = np.tile(mean_user.reshape((n_users, 1)), (1, n_movies))
    score = rmse(R12_true[hidden], R12_pred[hidden])
    print('RMSE(mean user): {}'.format(score))

    # mean movie
    R12_pred = np.tile(mean_movie.reshape((1, n_movies)), (n_users, 1))
    score = rmse(R12_true[hidden], R12_pred[hidden])
    print('RMSE(mean movie): {}'.format(score))

    # DFMF on ratings data only (it benefits if unknown values are set to mean)
    scores = []
    for _ in range(10):
        dfmf_fuser = skf.Dfmf(max_iter=100, fill_value=0,
                                 init_type='random')
        dfmf_mod = dfmf_fuser.fuse(graph_small)
        R12_pred = dfmf_mod.complete(graph_small['User ratings'])
        # R12_pred = scale(R12_pred, 0, 1)
        R12_pred += np.tile(mean_user.reshape((n_users, 1)), (1, n_movies))
        R12_pred += np.tile(mean_movie.reshape((1, n_movies)), (n_users, 1))
        R12_pred = scale(R12_pred, 0, 1)
        scores.append(rmse(R12_true[hidden], R12_pred[hidden]))
    print('RMSE(ratings; out-of-sample dfmf): {}'.format(np.mean(scores)))

    # DFMF (it benefits if unknown values are set to mean)
    scores = []
    for _ in range(10):
        dfmf_fuser = skf.Dfmf(max_iter=100, fill_value=0,
                                 init_type='random')
        dfmf_mod = dfmf_fuser.fuse(graph)
        R12_pred = dfmf_mod.complete(graph['User ratings'])
        # R12_pred = scale(R12_pred, 0, 1)
        R12_pred += np.tile(mean_user.reshape((n_users, 1)), (1, n_movies))
        R12_pred += np.tile(mean_movie.reshape((1, n_movies)), (n_users, 1))
        R12_pred = scale(R12_pred, 0, 1)
        scores.append(rmse(R12_true[hidden], R12_pred[hidden]))
    print('RMSE(out-of-sample dfmf): {}'.format(np.mean(scores)))

    # DFMC on ratings data only (proper treatment of unknown values)
    scores = []
    for _ in range(10):
        dfmc_fuser = skf.Dfmc(max_iter=100, init_type='random')
        dfmc_mod = dfmc_fuser.fuse(graph_small)
        R12_pred = dfmc_mod.complete(graph_small['User ratings'])
        R12_pred = scale(R12_pred, 0, 1)
        R12_pred += np.tile(mean_user.reshape((n_users, 1)), (1, n_movies))
        R12_pred += np.tile(mean_movie.reshape((1, n_movies)), (n_users, 1))
        R12_pred = scale(R12_pred, 0, 1)
        scores.append(rmse(R12_true[hidden], R12_pred[hidden]))
    print('RMSE(ratings; out-of-sample dfmc): {}'.format(np.mean(scores)))

    # DFMC (proper treatment of unknown values)
    scores = []
    for _ in range(10):
        dfmc_fuser = skf.Dfmc(max_iter=100, init_type='random')
        dfmc_mod = dfmc_fuser.fuse(graph)
        R12_pred = dfmc_mod.complete(graph['User ratings'])
        R12_pred = scale(R12_pred, 0, 1)
        R12_pred += np.tile(mean_user.reshape((n_users, 1)), (1, n_movies))
        R12_pred += np.tile(mean_movie.reshape((1, n_movies)), (n_users, 1))
        R12_pred = scale(R12_pred, 0, 1)
        scores.append(rmse(R12_true[hidden], R12_pred[hidden]))
    print('RMSE(out-of-sample dfmc): {}'.format(np.mean(scores)))

    # in-sample error (should be very close to zero for large rank values)
    score = rmse(R12_true[~hidden], R12_pred[~hidden])
    print('RMSE(in-sample dfmc): {}'.format(score))

    # PCA (requires scikit-learn module)
    from sklearn.decomposition import RandomizedPCA
    model = RandomizedPCA(n_components=10)
    R12 = graph['User ratings'].data.filled()
    pca_mod = model.fit(R12)
    R12_pred = scale(pca_mod.inverse_transform(pca_mod.transform(R12)), 0, 1)
    score = rmse(R12_true[hidden], R12_pred[hidden])
    print('RMSE(pca): {}'.format(score))

    # NMF (requires Nimfa module)
    import nimfa
    R12 = graph['User ratings'].data.filled()
    R12[R12 < 0] = 0
    lsnmf = nimfa.Lsnmf(R12, seed='random_vcol', rank=50, max_iter=100)
    lsnmf_fit = lsnmf()
    R12_pred = scale(np.dot(lsnmf_fit.basis(), lsnmf_fit.coef()), 0, 1)
    score = rmse(R12_true[hidden], R12_pred[hidden])
    print('RMSE(nmf): {}'.format(score))


if __name__ == "__main__":
    main()



import unittest

import numpy as np
from skfusion.fusion import Relation, ObjectType, \
    FusionGraph, Dfmf, DfmfTransform


class TestBase(unittest.TestCase):
    def test_pipeline(self):
        rnds = np.random.RandomState(0)
        R12 = rnds.rand(50, 30)
        R13 = rnds.rand(50, 40)
        R23 = rnds.rand(30, 40)

        t1 = ObjectType('type1', 30)
        t2 = ObjectType('type2', 40)
        t3 = ObjectType('type3', 40)
        relations = [Relation(R12, t1, t2),
                     Relation(R13, t1, t3), Relation(R23, t2, t3)]
        fusion_graph = FusionGraph()
        fusion_graph.add_relations_from(relations)

        fuser = Dfmf(random_state=rnds).fuse(fusion_graph)
        self.assertEqual(fuser.factor(t1).shape, (50, 30))
        self.assertEqual(fuser.factor(t2).shape, (30, 40))
        self.assertEqual(fuser.factor(t3).shape, (40, 40))
        self.assertEqual(fuser.backbone(relations[0]).shape, (30, 40))
        self.assertEqual(fuser.backbone(relations[1]).shape, (30, 40))
        self.assertEqual(fuser.backbone(relations[2]).shape, (40, 40))

        new_R12 = rnds.rand(15, 30)
        new_R13 = rnds.rand(15, 40)

        new_relations = [Relation(new_R12, t1, t2), Relation(new_R13, t1, t3)]
        new_graph = FusionGraph(new_relations)

        transformer = DfmfTransform(random_state=rnds).transform(t1, new_graph, fuser)
        self.assertEqual(transformer.factor(t1).shape, (15, 30))


if __name__ == "__main__":
    unittest.main()



"""
==========================================================================
Fusion of three data sources for gene function prediction in Dictyostelium
==========================================================================

Fuse three data sets: gene expression data (Miranda et al., 2013, PLoS One),
slim gene annotations from Gene Ontology and protein-protein interaction
network from STRING database.

Learnt latent matrix factors are utilized for the prediction of slim GO
terms in Dictyostelium genes that are unavailable in the training phase.

This example demonstrates how predictive performance can
be improved if prediction model relies on features extracted by data fusion
instead of on raw data.
"""
print(__doc__)

from sklearn import cross_validation, ensemble, metrics
import numpy as np

from skfusion import datasets
from skfusion import fusion as skf


def mf(train_idx, test_idx, term_idx):
    ann = dicty[gene][go_term][0].data.copy()
    ann[test_idx, :] = 0
    relations = [
        skf.Relation(ann, gene, go_term),
        skf.Relation(dicty[gene][exp_cond][0].data, gene, exp_cond),
        skf.Relation(dicty[gene][gene][0].data, gene, gene)]
    fusion_graph = skf.FusionGraph(relations)

    fuser = skf.Dfmf(max_iter=10, n_run=1, init_type="random_vcol", random_state=0)

    p = 0.7
    gene.rank = p * dicty[gene][go_term][0].data.shape[0]
    exp_cond.rank = p * dicty[gene][exp_cond][0].data.shape[1]
    go_term.rank = p * dicty[gene][go_term][0].data.shape[1]
    fuser.fuse(fusion_graph)
    X = fuser.complete(fusion_graph[gene][exp_cond][0])

    X_train = X[train_idx, :]
    y_train = dicty[gene][go_term][0].data[train_idx, term_idx]
    clf = ensemble.RandomForestClassifier(n_estimators=200)
    clf.fit(X_train, y_train)
    X_new = X[test_idx, :]
    y_pred = clf.predict_proba(X_new)[:, 1]
    return y_pred


def rf(train_idx, test_idx, term_idx):
    X_train = dicty[gene][exp_cond][0].data[train_idx, :]
    y_train = dicty[gene][go_term][0].data[train_idx, term_idx]
    clf = ensemble.RandomForestClassifier(n_estimators=200)
    clf.fit(X_train, y_train)
    X_new = dicty[gene][exp_cond][0].data[test_idx, :]
    y_pred = clf.predict_proba(X_new)[:, 1]
    return y_pred


def main():
    n_folds = 10
    n_genes, n_terms = dicty[gene][go_term][0].data.shape

    for t, term_idx in enumerate(range(n_terms)):
        term = dicty[gene][go_term][0].col_names[term_idx]
        print("Term: %s" % term)
        y_true = dicty[gene][go_term][0].data[:, term_idx]

        cls_size = int(y_true.sum())
        if cls_size > n_genes - 20 or cls_size < 20:
            continue

        cv = cross_validation.StratifiedKFold(y_true, n_folds=n_folds)
        y_pred_mf = np.zeros_like(y_true)
        y_pred_rf = np.zeros_like(y_true)
        for i, (train_idx, test_idx) in enumerate(cv):
            print("\tFold: %d" % (i+1))
            # Let"s make predictions from fused data representation
            y_pred_mf[test_idx] = mf(train_idx, test_idx, term_idx)
            # Let"s make predictions from raw data
            y_pred_rf[test_idx] = rf(train_idx, test_idx, term_idx)

        mfa = metrics.roc_auc_score(y_true, y_pred_mf)
        rfa = metrics.roc_auc_score(y_true, y_pred_rf)
        print("(%2d/%2d): %10s MF: %0.3f RF: %0.3f" % (t+1, n_terms, term, mfa, rfa))


if __name__ == "__main__":
    dicty = datasets.load_dicty()
    gene = dicty.get_object_type("Gene")
    go_term = dicty.get_object_type("GO term")
    exp_cond = dicty.get_object_type("Experimental condition")

    main()



from collections import defaultdict
from abc import ABCMeta, abstractmethod

import numpy as np


__all__ = ['FusionBase', 'FusionFit', 'FusionTransform', 'DataFusionError']


class FusionBase(object):
    """Base class for data fusion.

    Warning: This class should not be used directly.
    Use derived classes instead.

    Attributes
    ----------
    factors_ :
    backbones_ :
    """
    __metaclass__ = ABCMeta
    _params = None

    @abstractmethod
    def __init__(self):
        self.factors_ = defaultdict(list)
        self.backbones_ = defaultdict(list)

    def _set_params(self, values):
        self._params = values
        # first argument is 'self'
        del self._params['self']
        del self._params['__class__']
        self.__dict__.update(self._params)

    def factor(self, object_type, run=None):
        """Return fused latent matrix factor of an object type.

        Parameters
        ----------
        object_type :
        run :

        Returns
        -------
        G :
        """
        if object_type not in self.fusion_graph.object_types:
            raise DataFusionError("Object type %s is not included "
                                  "in the fusion scheme" % object_type.name)
        if object_type not in self.factors_:
            raise DataFusionError("Unknown object type.")
        if self.n_run > 1 and run is None:
            return self._factor_iter(object_type)
        else:
            run = 0 if run is None else run
            return self.factors_[object_type][run]

    def _factor_iter(self, object_type):
        """Return an iterator over latent matrix factors from
        different runs.

        Parameters
        ----------
        object_type :
        """
        for run in range(self.n_run):
            yield self.factors_[object_type][run]

    def chain(self, row_type, col_type):
        """Express objects of type ``row_type`` in the fused latent space of
        ``col_type`` by propagating latent factors over the fusion graph.

        Parameters
        ----------
        row_type :
        col_type :

        Returns
        -------
        chains :
        """
        paths = [[row_type]]
        if row_type == col_type:
            yield paths[0]
        while paths:
            paths_new = []
            for path in paths:
                expand = [ot for ot in self.fusion_graph.out_neighbors(path[-1])
                          if ot not in path]
                refined_paths = [path + [ot] for ot in expand]
                for refined in refined_paths:
                    if refined[-1] == col_type:
                        yield refined
                    else:
                        paths_new.append(refined)
            paths = paths_new

    def __str__(self):
        pparams = ', '.join('{}={}'.format(k, v) for k, v in self._params.items())
        return '{}({})'.format(self.__class__.__name__, pparams)

    def __repr__(self):
        pparams = ', '.join('{}={}'.format(k, v) for k, v in self._params.items())
        return '{}({})'.format(self.__class__.__name__, pparams)


class FusionFit(FusionBase):
    """Base class for fused data space estimation.

    Warning: This class should not be used directly.
    Use derived classes instead.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self):
        super(FusionFit, self).__init__()

    def complete(self, relation, run=None):
        """Return reconstructed relation matrix.

        Parameters
        ----------
        relation :
        run :

        Returns
        -------
        R12_hat : array-like, shape (n_obj1, n_obj2)
        """
        if relation.row_type not in self.fusion_graph.object_types or \
            relation.col_type not in self.fusion_graph.object_types:
            raise DataFusionError("Object type %s or %s are not included "
                                  "in the fusion scheme" % (
                relation.row_type.name, relation.col_type.name))
        if self.n_run > 1 and run is None:
            return self._complete_iter(relation)
        else:
            run = 0 if run is None else run
            G1 = self.factor(relation.row_type, run)
            S12 = self.backbone(relation, run)
            G2 = self.factor(relation.col_type, run)
            R12_hat = np.dot(G1, np.dot(S12, G2.T))
            if relation.postprocessor:
                R12_hat = relation.postprocessor(R12_hat)
            return R12_hat

    def _complete_iter(self, relation):
        """Return an iterator over completed relation matrix obtained
        from different runs

        Parameters
        ----------
        relation :
        run :

        Returns
        -------
        """
        for run in range(self.n_run):
            G1 = self.factor(relation.row_type, run)
            S12 = self.backbone(relation, run)
            G2 = self.factor(relation.col_type, run)
            R12_hat = np.dot(G1, np.dot(S12, G2.T))
            if relation.postprocessor:
                R12_hat = relation.postprocessor(R12_hat)
            yield R12_hat

    def backbone(self, relation, run=None):
        """Return fused latent matrix factor of a relation.

        Parameters
        ----------
        relation :
        run :

        Returns
        -------
        """
        if relation.row_type not in self.fusion_graph.object_types \
                or relation.col_type not in self.fusion_graph.object_types:
            raise DataFusionError('Object types are not recognized.')
        if relation not in self.backbones_:
            raise DataFusionError("Unknown relation.")
        if self.n_run > 1 and run is None:
            return self._backbone_iter(relation)
        else:
            run = 0 if run is None else run
            return self.backbones_[relation][run]

    def _backbone_iter(self, relation):
        """Return an iterator over backbone latent matrix factors from
        different runs.

        Parameters
        ----------
        relation :

        Returns
        -------
        """
        for run in range(self.n_run):
            yield self.backbones_[relation][run]


class FusionTransform(FusionBase):
    """Base class for online data transformation into the fused space

    Warning: This class should not be used directly.
    Use derived classes instead.

    Attributes
    ----------
    target :
    fusion_graph :
    fuser :
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self):
        super(FusionTransform, self).__init__()

    def _validate_graph(self):
        if self.target not in self.fusion_graph.object_types:
            raise DataFusionError("Object type %s is not included " \
                                  "in the fusion scheme." % self.target.name)
        for relation in self.fusion_graph.relations:
            if self.target not in [relation.row_type, relation.col_type]:
                raise DataFusionError("Relation must include target "
                                  "object type: %s." % self.target.name)

    def chain(self, row_type=None, col_type=None):
        """Express objects of type ``target`` in the fused latent space of
        ``col_type`` by propagating latent factors over the fusion graph.

        Parameters
        ----------
        row_type :
        col_type :
        """
        if row_type is not None and col_type is not None and \
                        row_type is not self.target:
            raise DataFusionError("Starting type should be target type: "
                                  "%s" % self.target.name)
        col_type = row_type if col_type is None else col_type
        return FusionBase.chain(self, self.target, col_type)


class DataFusionError(Exception):
    pass



"""
=========================================
Data Fusion by Matrix completion (`dfmc`)
=========================================
"""

import logging
from operator import add
from collections import defaultdict
from functools import reduce

import numpy as np
import scipy.linalg as spla
from joblib import Parallel, delayed

from ._init import initialize


def __bdot(A, B, i, j, obj_types):
    entry = []
    if isinstance(list(A.values())[0], list):
        for l in range(len(A.get((i, j), []))):
            ll = [np.dot(A[i, k][l], B[k, j]) for k in obj_types
                  if (i, k) in A and (k, j) in B]
            if len(ll) > 0:
                tmp = reduce(add, ll)
                entry.append(np.nan_to_num(tmp))
    elif isinstance(list(B.values())[0], list):
        for l in range(len(B.get((i, j), []))):
            ll = [np.dot(A[i, k], B[k, j][l]) for k in obj_types
                  if (i, k) in A and (k, j) in B]
            if len(ll) > 0:
                tmp = reduce(add, ll)
                entry.append(np.nan_to_num(tmp))
    else:
        ll = [np.dot(A[i, k], B[k, j]) for k in obj_types
              if (i, k) in A and (k, j) in B]
        if len(ll) > 0:
            entry = reduce(add, ll)
            entry = np.nan_to_num(entry)
    return i, j, entry


def _par_bdot(A, B, obj_types, verbose, n_jobs):
    """Parallel block matrix multiplication.

    Parameters
    ----------
    A : dictionary of array-like objects
        Block matrix.

    B : dictionary of array-like objects
        Block matrix.

    obj_types : array-like
        Identifiers of object types.

    verbose : int
         The amount of verbosity.

    n_jobs: int (default=1)
        Number of jobs to run in parallel

    Returns
    -------
    C : dictionary of array-like objects
        Matrix product, A*B.
    """
    parallelizer = Parallel(n_jobs=n_jobs, max_nbytes=1e3, verbose=verbose,
                            backend="multiprocessing")
    task_iter = (delayed(__bdot)(A, B, i, j, obj_types)
                 for i in obj_types for j in obj_types)
    entries = parallelizer(task_iter)
    C = {(i, j): entry for i, j, entry in entries if entry != []}
    return C


def _transpose(A):
    """Block matrix transpose.

    Parameters
    ----------
    A : dictionary of array-like objects
        Block matrix.

    Returns
    -------
    At : dictionary of array-like objects
        Block matrix with each of its block transposed.
    """
    At = {k: V.T for k, V in A.items()}
    return At


def count_objects(obj_types, R):
    """Count objects of each object type.

    Parameters
    ----------
    obj_types : set-like
        Object types

    R : dictionary of array-like objects
        Relation matrices

    Returns
    -------
    obj_type2n_obj : dictionary of int
        Number of objects per object type
    """
    obj_type2n_obj = {}
    for r in R:
        i, j = r
        for l in range(len(R[r])):
            for ax, obj_type in enumerate([i, j]):
                ni = obj_type2n_obj.get(obj_type, R[i,j][l].shape[ax])
                if ni != R[i,j][l].shape[ax]:
                    logging.critical("Relation matrix R_(%s,%s) dimension "
                                     "mismatch" % (i, j))
                obj_type2n_obj[obj_type] = ni

    if set(obj_types) != set(obj_type2n_obj.keys()):
        logging.critical("Object type specification mismatch")
    return obj_type2n_obj


def _update_G_for_Rij(Rij_l, G_i, G_j, Sij_l):
    """Multiplicative update of latent factors G_i and G_j
    due to l-th relation matrix R_ij.

    Parameters
    ----------
    Rij_l :
        The l-th relation matrix for R_ij

    G_i :  array-like, shape (n_objects_i, n_latent_i)
        Current estimate of latent factor G_i

    G_j : array-like, shape (n_objects_j, n_latent_j)
        Current estimate of latent factor G_j

    Sij_l : array-like, shape(n_latent_i, n_latent_j)
        Current estimation of the l-th latent factor S_ij

    Returns
    -------
    G_enum_i :
    G_denom_i :
    G_enum_j :
    G_denom_j :
    """
    tmp1 = np.dot(Rij_l, np.dot(G_j, Sij_l.T))
    t = tmp1 > 0
    tmp1p = np.multiply(t, tmp1)
    tmp1n = np.multiply(t-1, tmp1)

    tmp2 = np.dot(Sij_l, np.dot(G_j.T, np.dot(G_j, Sij_l.T)))
    t = tmp2 > 0
    tmp2p = np.multiply(t, tmp2)
    tmp2n = np.multiply(t-1, tmp2)

    tmp4 = np.dot(Rij_l.T, np.dot(G_i, Sij_l))
    t = tmp4 > 0
    tmp4p = np.multiply(t, tmp4)
    tmp4n = np.multiply(t-1, tmp4)

    tmp5 = np.dot(Sij_l.T, np.dot(G_i.T, np.dot(G_i, Sij_l)))
    t = tmp5 > 0
    tmp5p = np.multiply(t, tmp5)
    tmp5n = np.multiply(t-1, tmp5)

    G_enum_i = tmp1p + np.dot(G_i, tmp2n)
    G_denom_i = tmp1n + np.dot(G_i, tmp2p)

    G_enum_j = tmp4p + np.dot(G_j, tmp5n)
    G_denom_j = tmp4n + np.dot(G_j, tmp5p)

    return (G_enum_i, G_denom_i), (G_enum_j, G_denom_j)


def dfmc(R, M, Theta, obj_types, obj_type2rank, max_iter=10,
         init_type="random_vcol", stopping=None, stopping_system=None,
         verbose=0, compute_err=False, callback=None, random_state=None,
         n_jobs=1):
    """Data fusion by matrix completion.

    Parameters
    ----------
    R : dictionary of array-like objects
        Relation matrices.

    M : dictionary of array-like objects
        Mask matrices

    Theta : dictionary of array-like objects
        Constraint matrices.

    obj_types : array-like
        Identifiers of object types.

    obj_type2rank : dict-like
        Factorization ranks of object types.

    max_iter : int, optional (default=10)
        Maximum number of iterations to perform.

    init_type : string, optional (default="random_vcol")
        The algorithm to initialize latent matrix factors.

    stopping : tuple (target_matrix, eps), optional (default=None)
        Stopping criterion. Terminate iteration if the reconstruction
        error of target matrix improves by less than eps.

    stopping_system : float, optional (default=None)
        Stopping criterion. Terminate iteration if the reconstruction
        error of the fused system improves by less than eps. compute_err is
        to True to compute the error of the fused system.

    verbose : int, optional (default=0)
         The amount of verbosity. Larger value indicates greater verbosity.

    compute_err : bool, optional (default=False)
        Compute the reconstruction error of every relation matrix if True.

    callback : callable, optional
        An optional user-supplied function to call after each iteration. Called
        as callback(G, S, cur_iter), where S and G are current latent estimates.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance
        used by np.random.

    n_jobs: int (default=1)
        Number of jobs to run in parallel

    Returns
    -------
    G :
    S :
    """
    verbose1 = verbose
    verbose = 50 - verbose
    logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s",
                        datefmt="%m/%d/%Y %I:%M:%S %p", level=verbose)
    R_to_init = {k: V[0] for k, V in R.items()}
    obj_type2n_obj = count_objects(obj_types, R)
    G = initialize(obj_types, obj_type2n_obj, obj_type2rank, R_to_init,
                   init_type, random_state)
    S = None

    if stopping:
        err_target = (None, None)
    if stopping_system:
        err_system = (None, None)
        compute_err = True

    logging.info("Solving for Theta_p and Theta_n")
    Theta_p, Theta_n = defaultdict(list), defaultdict(list)
    for r, thetas in Theta.items():
        for theta in thetas:
            t = theta > 0
            Theta_p[r].append(np.multiply(t, theta))
            Theta_n[r].append(np.multiply(t-1, theta))

    # For reasons of matrix completion the data should be copied inside DFMC
    R = {r: [m.copy() for m in data] for r, data in R.items()}

    obj = []
    for iter in range(max_iter):
        if iter > 1 and stopping and err_target[1]-err_target[0] < stopping[1]:
            logging.info("Early stopping: target matrix change < %5.4f" \
                      % stopping[1])
            break
        if iter > 1 and stopping_system and \
                                err_system[1]-err_system[0] < stopping_system:
            logging.info("Early stopping: matrix system change < %5.4f" \
                      % stopping_system)
            break

        logging.info("Completion iteration: %d" % iter)

        #######################################################################
        ########################## Matrix Completion ##########################

        if iter == 0:
            for r in M:
                for l in range(len(R[r])):
                    if M[r][l] is None:
                        continue
                    R[r][l][M[r][l]] = 0.

        #######################################################################
        ########################### General Update ############################

        pGtG = {}
        for r in G:
            logging.info("Computing GrtGr: %s" % str(r))
            GrtGr = np.nan_to_num(np.dot(G[r].T, G[r]))
            # numpy.linalg.pinv approximates the Moore-Penrose pseudo inverse
            # using SVD (the lapack method dgesdd), whereas scipy.linalg.pinv
            # solves a linear system in the least squares sense (the lapack
            # method dgelss). Alternatively, numpy.linalg.solve solve a
            # linear system in the least squares sense. The latter turns out
            # to be unstable in this case.
            pGtG[r] = spla.pinv(GrtGr)

        logging.info("Start to update S")

        tmp1 = _par_bdot(G, pGtG, obj_types, verbose1, n_jobs)
        tmp2 = _par_bdot(R, tmp1, obj_types, verbose1, n_jobs)
        tmp3 = _par_bdot(_transpose(G), tmp2, obj_types, verbose1, n_jobs)
        S = _par_bdot(pGtG, tmp3, obj_types, verbose1, n_jobs)

        #######################################################################
        ########################## Matrix Completion ##########################

        for r in M:
            for l in range(len(M[r])):
                if M[r][l] is None:
                    continue
                i, j = r
                Rij_app = np.dot(G[i, i], np.dot(S[i, j][l], G[j, j].T))
                R[r][l][M[r][l]] = Rij_app[M[r][l]]

        #######################################################################
        ########################### General Update ############################

        logging.info("Start to update G")

        G_enum = {r: np.zeros(Gr.shape) for r, Gr in G.items()}
        G_denom = {r: np.zeros(Gr.shape) for r, Gr in G.items()}

        for r in R:
            i, j = r
            logging.info("Update G due to R_%s,%s" % (i, j))
            # Start parallel update
            logging.info("Start parallel update of R_%s,%s" % (i, j))

            parallelizer = Parallel(n_jobs=n_jobs, max_nbytes=1e3,
                                    backend="multiprocessing", verbose=verbose1)
            task_iter = (delayed(_update_G_for_Rij)(
                    R[i,j][l], G[i,i], G[j,j], S[i,j][l]) for l in range(len(R[r])))
            G_ij_update = parallelizer(task_iter)

            for (G_enum_i, G_denom_i), (G_enum_j, G_denom_j) in G_ij_update:
                G_enum[i, i] += G_enum_i
                G_denom[i, i] += G_denom_i

                G_enum[j, j] += G_enum_j
                G_denom[j, j] += G_denom_j

        logging.info("Update of G due to constraint matrices")
        for r, thetas_p in Theta_p.items():
            logging.info("Considering Theta pos. %s" % str(r))
            for theta_p in thetas_p:
                G_denom[r] += np.dot(theta_p, G[r])
        for r, thetas_n in Theta_n.items():
            logging.info("Considering Theta neg. %s" % str(r))
            for theta_n in thetas_n:
                G_enum[r] += np.dot(theta_n, G[r])

        for r in G:
            G[r] = np.multiply(G[r], np.sqrt(
                np.divide(G_enum[r], np.maximum(G_denom[r], np.finfo(np.float).eps))))

        #######################################################################

        if stopping:
            target, eps = stopping
            r, l = target
            err_target = (np.linalg.norm(R[r][l]-np.dot(G[r[0],r[0]],
                    np.dot(S[r][l], G[r[1],r[1]].T))), err_target[0])

        if compute_err:
            s = 0
            for r in R:
                i, j = r
                for l in range(len(R[r])):
                    Rij_app = np.dot(G[i, i], np.dot(S[r][l], G[j, j].T))
                    r_err = np.linalg.norm(R[r][l]-Rij_app, "fro")
                    logging.info("Relation %s^(%d) norm difference: " \
                              "%5.4f" % (str(r), l, r_err))
                    s += r_err
            logging.info("Error (objective function value): %5.4f" % s)
            obj.append(s)
            if stopping_system:
                err_system = (s, err_system[0])

        if callback:
            callback(G, S, iter)

    if compute_err:
        logging.info("Violations of optimization objective: %d/%d " % (
            int(np.sum(np.diff(obj) > 0)), len(obj)))
    return G, S


if 0:
    # toy example for matrix completion
    rnds = np.random.RandomState(0)
    from sklearn import datasets

    R12 = datasets.load_digits(7).data
    obj_types = [0, 1, 2]
    obj_type2rank = {0: 100, 1: 200, 2:10}

    R12b = np.random.rand(R12.shape[0], R12.shape[1])
    R12b[R12b < 0.3] = 0

    # mask R12
    grid = np.indices(R12.shape)
    idx = list(zip(grid[0].ravel(), grid[1].ravel()))
    rnds.shuffle(idx)
    idxi, idxj = list(zip(*idx[:int(0.9 * R12.size)]))
    mR12 = np.ma.array(R12)
    mR12[idxi, idxj] = np.ma.masked
    print("R12 mask: %d" % mR12.mask.sum())

    R23 = np.random.rand(R12.shape[1], 40)

    R = {(0,1): [R12, R12b], (1,2): [R23.copy(), R23]}
    Theta = {}
    M = {(0,1): [mR12.mask, mR12.mask.copy()]}

    ## target matrix
    i, j = 0, 1
    ##

    # Evaluate completion (mask)
    G, S = dfmc(R, M, Theta, obj_types, obj_type2rank,
                max_iter=10, init_type="random_vcol", stopping=None,
                stopping_system=None, verbose=0, compute_err=True,
                callback=None, random_state=rnds, n_jobs=1)

    Rij_app = np.dot(G[i,i], np.dot(S[i, j][0], G[j, j].T))
    X_pred = R[i,j][0][M[i,j][0]] - Rij_app[M[i,j][0]]
    oob_fro = np.sqrt(np.sum(X_pred**2) / X_pred.size)
    print("OOB(Completion) RMSE for R(%d, %d): %5.3f" % (i+1, j+1, oob_fro))

    rnd = R[i,j][0][M[i,j][0]]
    rnds.shuffle(rnd)
    X_rnd = R[i,j][0][M[i,j][0]] - rnd
    rnd_fro = np.sqrt(np.sum(X_rnd**2) / X_rnd.size)
    print("Rnd RMSE for R(%d, %d): %5.3f" % (i+1, j+1, rnd_fro))


    print("R12 set to zero: %d" % np.size(R12[idxi, idxj]))

    O = {(0,1): [R12.copy(), R12b.copy()], (1,2): [R23.copy()]}
    M = {}
    R12[idxi, idxj] = 0
    R = {(0,1): [R12, R12b], (1,2): [R23]}

    def cb(G, S, cur_iter):
        print('Iteration: %d' % cur_iter)

    # Evaluate factorization (no mask)
    G, S = dfmc(R, M, Theta, obj_types, obj_type2rank,
                max_iter=10, init_type="random_vcol", stopping=None,
                stopping_system=None, verbose=0, compute_err=False,
                callback=cb, random_state=rnds, n_jobs=1)

    Rij_app = np.dot(G[i,i], np.dot(S[i, j][0], G[j, j].T))
    X_pred = O[i,j][0][idxi, idxj] - Rij_app[idxi, idxj]
    oob_fro = np.sqrt(np.sum(X_pred**2) / X_pred.size)
    print("OOB(Factorization) RMSE for R(%d, %d): %5.3f" % (i+1, j+1, oob_fro))



from itertools import product
from collections import defaultdict

import numpy as np
from joblib import Parallel, delayed

from ..base import FusionFit, FusionTransform
from ._dfmf import dfmf, transform


__all__ = ['Dfmf', 'DfmfTransform']


def parallel_dfmf_wrapper(**params):
    return dfmf(**params)


class Dfmf(FusionFit):
    """Data fusion by matrix factorization.

    Attributes
    ---------
    fusion_graph :
    max_iter :
    init_type :
    n_run :
    stopping :
    stopping_system :
    fill_value :
    verbose :
    compute_err :
    callback :
    random_state :
    n_jobs :

    Parameters
    ----------
    max_iter :
    init_type :
    n_run :
    stopping :
    stopping_system :
    fill_value :
    verbose :
    compute_err :
    callback :
    random_state :
    n_jobs :
    """
    def __init__(self, max_iter=100, init_type='random_c', n_run=1,
                 stopping=None, stopping_system=None, fill_value=0, verbose=0,
                 compute_err=False, callback=None, random_state=None, n_jobs=1):
        super(Dfmf, self).__init__()
        self._set_params(vars())

    def fuse(self, fusion_graph):
        """Run data fusion factorization algorithm.

        Parameters
        ----------
        fusion_graph :
        """
        self.fusion_graph = fusion_graph
        if not isinstance(self.random_state, np.random.RandomState):
            self.random_state = np.random.RandomState(self.random_state)

        object_types = set([ot for ot in self.fusion_graph.object_types])
        object_type2rank = {ot: int(ot.rank) for ot in self.fusion_graph.object_types}

        R, T = {}, {}
        for row_type, col_type in product(self.fusion_graph.object_types, repeat=2):
            for relation in self.fusion_graph.get_relations(row_type, col_type):
                if relation.preprocessor:
                    data = relation.preprocessor(relation.data)
                else:
                    data = relation.data
                if np.ma.is_masked(data):
                    data.fill_value = self.fill_value
                    data = data.filled()
                data[~np.isfinite(data)] = self.fill_value
                X = R if relation.row_type != relation.col_type else T
                X[relation.row_type, relation.col_type] = X.get((
                    relation.row_type, relation.col_type), [])
                X[relation.row_type, relation.col_type].append(data)

        parallelizer = Parallel(n_jobs=self.n_jobs, max_nbytes=1e3, verbose=self.verbose)
        task_iter = (delayed(parallel_dfmf_wrapper)(
            R=R, Theta=T, obj_types=object_types, obj_type2rank=object_type2rank,
            max_iter=self.max_iter, init_type=self.init_type, stopping=self.stopping,
            stopping_system=self.stopping_system, verbose=self.verbose,
            compute_err=self.compute_err, callback=self.callback, random_state=self.random_state,
            n_jobs=self.n_jobs)
                     for _ in range(self.n_run))
        entries = parallelizer(task_iter)

        self.factors_ = defaultdict(list)
        self.backbones_ = defaultdict(list)
        for G, S in entries:
            for (object_type, _), factor in G.items():
                self.factors_[object_type].append(factor)

            for (row_type, col_type), backbones in S.items():
                for i, relation in enumerate(self.fusion_graph.get_relations(row_type, col_type)):
                    self.backbones_[relation].append(backbones[i])
        return self


def parallel_dfmf_transform_wrapper(fuser, run, **params):
    G = {(object_type, object_type): fuser.factor(object_type, run)
         for object_type in fuser.fusion_graph.object_types}
    S = {(relation.row_type, relation.col_type): [fuser.backbone(relation, run)]
         for relation in fuser.fusion_graph.relations
         if relation.row_type != relation.col_type}
    return transform(G=G, S=S, **params)


class DfmfTransform(FusionTransform):
    """Online data transformer into fused space.

    Attributes
    ----------
    target :
    fusion_graph :
    fuser :
    max_iter :
    init_type :
    stopping :
    stopping_system :
    fill_value :
    verbose :
    compute_err :
    random_state :
    n_jobs :

    Parameters
    ----------
    fuser :
    max_iter :
    init_type :
    stopping :
    stopping_system :
    fill_value :
    verbose :
    compute_err :
    random_state :
    n_jobs :
    """
    def __init__(self, max_iter=100, init_type=None, n_run=1, stopping=None,
                 stopping_system=None, fill_value=0, verbose=0, compute_err=False,
                 callback=None, random_state=None, n_jobs=1):
        super(DfmfTransform, self).__init__()
        self._set_params(vars())

    def transform(self, target, fusion_graph, fuser):
        """Transform the data into the space given by Fuser.

        Parameters
        ----------
        target :
        fusion_graph :
        fuser :
        """
        self.target = target
        self.fusion_graph = fusion_graph
        self.fuser = fuser
        self._validate_graph()

        init_type = self.init_type if self.init_type is not None else self.fuser.init_type
        if not isinstance(self.random_state, np.random.RandomState):
            self.random_state = np.random.RandomState(self.random_state)

        object_type2rank = {ot: int(ot.rank) for ot in self.fusion_graph.object_types}

        R, T = {}, {}
        for row_type, col_type in product(self.fusion_graph.object_types, repeat=2):
            for relation in self.fusion_graph.get_relations(row_type, col_type):
                if relation.preprocessor:
                    data = relation.preprocessor(relation.data)
                else:
                    data = relation.data
                if np.ma.is_masked(data):
                    data.fill_value = self.fill_value
                    data = data.filled()
                data[~np.isfinite(data)] = self.fill_value
                X = R if relation.row_type != relation.col_type else T
                X[relation.row_type, relation.col_type] = X.get((
                    relation.row_type, relation.col_type), [])
                X[relation.row_type, relation.col_type].append(data)

        parallelizer = Parallel(n_jobs=self.n_jobs, max_nbytes=1e3, verbose=self.verbose)
        task_iter = (delayed(parallel_dfmf_transform_wrapper)(
            self.fuser, run,
            R_ij=R, Theta_i=T, target_obj_type=self.target,
            obj_type2rank=object_type2rank, max_iter=self.max_iter, init_type=init_type,
            stopping=self.stopping, stopping_system=self.stopping_system, verbose=self.verbose,
            compute_err=self.compute_err, callback=self.callback, random_state=self.random_state)
                     for run in range(self.n_run))
        entries = parallelizer(task_iter)

        self.factors_ = defaultdict(list)
        for G_new in entries:
            self.factors_[self.target].append(G_new)
        return self



#! /usr/bin/env python
descr = """A Python module for data fusion built on top of factorized models."""

from setuptools import setup, find_packages

DISTNAME = 'scikit-fusion'
DESCRIPTION = 'A Python module for data fusion built on top of factorized models.'
LONG_DESCRIPTION = open('README.rst').read()
MAINTAINER = 'Marinka Zitnik'
MAINTAINER_EMAIL = 'marinka.zitnik@fri.uni-lj.si'
URL = 'http://xyz.biolab.si'
LICENSE = 'GPLv3'
VERSION = '0.1'

INSTALL_REQUIRES = (
    'numpy>=1.7.0',
    'scipy>=0.12.0',
    'joblib>=0.8.4',
)


def setup_package():
    setup(name=DISTNAME,
          maintainer=MAINTAINER,
          maintainer_email=MAINTAINER_EMAIL,
          description=DESCRIPTION,
          license=LICENSE,
          url=URL,
          version=VERSION,
          long_description=LONG_DESCRIPTION,
          classifiers=['Intended Audience :: Science/Research',
                       'Intended Audience :: Developers',
                       'License :: OSI Approved',
                       'Programming Language :: Python',
                       'Topic :: Software Development',
                       'Topic :: Scientific/Engineering',
                       'Topic :: Scientific/Engineering :: Artificial Intelligence',
                       'Topic :: Scientific/Engineering :: Bio-Informatics',
                       'Programming Language :: Python :: 2',
                       'Programming Language :: Python :: 3',
                       ],
          install_requires=INSTALL_REQUIRES,
          packages=find_packages(),
          package_data={
              'skfusion.datasets': ['data/*/*']
          },
          test_suite='skfusion.tests.test_suite',
)


if __name__ == "__main__":
    setup_package()



from operator import itemgetter

import numpy as np


def initialize(obj_types, obj_type2n_obj, obj_type2rank, R, init_typ, random_state):
    init_types = {"random": _random, "random_c": _random_c, "random_vcol": _random_vcol}
    return init_types[init_typ](obj_types, obj_type2n_obj, obj_type2rank, R, random_state)


def _random(obj_types, obj_type2n_obj, obj_type2rank, R, random_state):
    G = {}
    for obj_type in obj_types:
        ni = obj_type2n_obj[obj_type]
        ci = obj_type2rank[obj_type]
        G[obj_type, obj_type] = random_state.rand(ni, ci)
    return G


def _random_c(obj_types, obj_type2n_obj, obj_type2rank, R, random_state):
    G = {}
    for obj_type in obj_types:
        ci = obj_type2rank[obj_type]
        G[obj_type, obj_type] = 1e-5 * np.ones((obj_type2n_obj[obj_type], ci))

        for obj_types, R12 in R.items():
            if obj_type not in obj_types:
                continue
            Rij = R12 if obj_type == obj_types[0] else R12.T
            p_c = int(.2 * Rij.shape[1])
            l_c = int(.5 * Rij.shape[1])
            cols_norm = [np.linalg.norm(Rij[:,i], 2) for i in range(Rij.shape[1])]
            top_c = sorted(enumerate(cols_norm), key=itemgetter(1), reverse=True)[:l_c]
            top_c = list(list(zip(*top_c))[0])
            Gi = np.zeros(G[obj_type, obj_type].shape)
            for i in range(ci):
                random_state.shuffle(top_c)
                Gi[:,i] = Rij[:, top_c[:p_c]].mean(axis=1)
            G[obj_type, obj_type] += np.abs(Gi)

    return G


def _random_vcol(obj_types, obj_type2n_obj, obj_type2rank, R, random_state):
    G = {}
    for obj_type in obj_types:
        ci = obj_type2rank[obj_type]
        G[obj_type, obj_type] = 1e-5 * np.ones((obj_type2n_obj[obj_type], ci))

        for obj_types, R12 in R.items():
            if obj_type not in obj_types:
                continue
            Rij = R12 if obj_type == obj_types[0] else R12.T
            p_c = int(.2 * Rij.shape[1])
            Gi = np.zeros(G[obj_type, obj_type].shape)
            idx = np.arange(Rij.shape[1])
            for i in range(ci):
                random_state.shuffle(idx)
                Gi[:, i] = Rij[:, idx[:p_c]].mean(axis=1)
            G[obj_type, obj_type] += np.abs(Gi)
    return G



from collections import defaultdict, OrderedDict, Iterable
from uuid import uuid1 as uuid

import numpy as np

from .base import DataFusionError


__all__ = ['FusionGraph', 'Relation', 'ObjectType']


class FusionGraph(object):
    """Container object for data sets and object types.

    Parameters
    ----------
    relations :

    Attributes
    ----------
    adjacency_matrix
    relations:
    object_types :
    """
    def __init__(self, relations=()):
        self.adjacency_matrix = {}
        self.relations = OrderedDict()
        self.object_types = OrderedDict()
        self._name2relation = {}
        self._name2object_type = {}
        self.add_relations_from(relations)

    @property
    def n_relations(self):
        return len(self.relations)

    @property
    def n_object_types(self):
        return len(self.object_types)

    def __getitem__(self, key):
        return self.adjacency_matrix.get(key, self._name2relation.get(key, None))

    def __setitem__(self, key, value):
        self.adjacency_matrix[key] = value

    def draw_networkx(self, filename=None, ax=None, *args, **kwargs):
        """Draw the data fusion graph using NetworkX and Matplotlib.

        Parameters
        ----------
        filename : str or file-like object
            A filename to output to. If str, the extension implies the format.
            If file-like object, pass the desired `format` explicitly.
            If None, the plot is drawn to a Matplotlib Axes object (can be
            supplied as `ax` keyword argument).

        **kwargs : optional keyword arguments
            Passed to ``networkx.draw_networkx()`` (and, optionally,
            ``matplotlib.figure.Figure.savefig()``).
        """
        import networkx as nx

        if filename and not ax:
            from matplotlib.figure import Figure
            ax = Figure().add_subplot(111)

        G = nx.MultiDiGraph()
        G.add_nodes_from(o.name for o in self.object_types)

        ot2count = defaultdict(int)
        for relation in self.relations:
            ot1 = relation.row_type
            ot2 = relation.col_type
            ot2count[ot1, ot2] += 1
            if ot1 != ot2:
                label = (r'$<\mathbf{R}_{%s,%s}^%d>$' %
                         (ot1.name, ot2.name, ot2count[ot1, ot2]))
            else:
                label = (r'$<\mathbf{\Theta}_%s^%d>$' %
                         (ot1.name, ot2count[ot1, ot2]))
            G.add_edge(ot1.name, ot2.name, label=label)

        nx.draw_networkx(G, *args,
                         ax=ax,
                         node_size=3000,
                         node_color='white',
                         **kwargs)
        edge_labels = {(u, v): d['label'] for u, v, d in G.edges(data=True)}
        nx.draw_networkx_edge_labels(G, nx.spring_layout(G), edge_labels=edge_labels)

        if filename:
            ax.figure.savefig(filename, **kwargs)
        return G

    def draw_graphviz(self, *args, **kwargs):
        """Draw the data fusion graph using PyGraphviz and save it to a file.

        Parameters
        ----------
        graph_attr : dict
            Dict of Graphviz graph attributes
        node_attr : dict
            Dict of Graphviz node attributes
        edge_attr : dict
            Dict of Graphviz edge attributes

        *args, **kwargs : optional
            Passed to `pygraphviz.AGraph.draw()` method.
        """
        import pygraphviz as pgv
        G = pgv.AGraph(strict=False, directed=True)
        # From http://graphviz.org/content/attrs
        G.graph_attr.update({
            'outputorder': 'edgesfirst',
            'packmode': 'graph',
            'pad': .3,
        }, **kwargs.pop('graph_attr', {}))
        G.node_attr.update({
            'fontsize': 11,
            'fontname': 'sans-serif',
            'fillcolor': 'white',
            'style': 'filled',
        }, **kwargs.pop('node_attr', {}))
        G.edge_attr.update({
            'fontsize': 9,
            'fontname': 'sans-serif',
        }, **kwargs.pop('edge_attr', {}))

        smallsize = .8 * float(G.node_attr['fontsize'])
        n_objects = {}
        for ot in self.object_types:
            # The maximum number of objects of this type featured in any of the
            # relations
            n = max(max([rel.data.shape[0] for rel in self.out_relations(ot)], default=0),
                    max([rel.data.shape[1] for rel in self.in_relations(ot)],  default=0))
            n_objects[ot] = n
            G.add_node(ot.name,
                       # This is relied upon by biolab/orange3; if you change this id,
                       # please let them know:
                       id='node `%s`' % ot.name,
                       label=('<%s<br/><font point-size="%.1f" color="grey">'
                               '%d</font>>' % (ot.name, smallsize, n)))
        relations = defaultdict(list)
        for rel in self.relations:
            relations[(rel.row_type, rel.col_type)].append(rel)
        for (ot1, ot2), rels in relations.items():
            label = (',<br/>&nbsp;'.join(rel.name for rel in rels if rel.name) or
                     '<b>%s</b>' % ('R' if ot1 != ot2 else '&Theta;'))
            label = '<&nbsp;' + label + '>'
            tooltip = ', '.join('[%d×%d]' % rel.data.shape for rel in rels)
            # Penwidth is normalized as the sum of relations' (defined) areas
            # divided by the largest possible area of the given two object types
            weight = sum(np.ma.count(rel.data) / n_objects[ot1] / n_objects[ot2]
                         for rel in rels)
            penwidth = np.clip(1.3 * weight, .5, 3)
            G.add_edge(ot1.name, ot2.name,
                       # This is relied upon by biolab/orange3; if you change this id,
                       # please let them know
                       id='edge `%s`->`%s`' % (ot1.name, ot2.name),
                       label=label,
                       tooltip=tooltip,
                       labelaligned=True,  # http://www.graphviz.org/content/allign-edge-labels-fit-its-path-svg-output
                       penwidth=penwidth)

        if len(args) < 3 and 'prog' not in kwargs:
            kwargs['prog'] = 'dot'
        G.draw(*args, **kwargs)

    def add_relation(self, relation):
        """Add a single relation to the fusion graph.

        Parameters
        ----------
        relation :
        """
        self.relations[relation] = True
        if relation.name:
            self._name2relation[relation.name] = relation
        self.object_types[relation.row_type] = True
        self.object_types[relation.col_type] = True
        self._name2object_type[relation.row_type.name] = relation.row_type
        self._name2object_type[relation.col_type.name] = relation.col_type
        neighbors = self.adjacency_matrix.get(relation.row_type, {})
        nbs_list = neighbors.get(relation.col_type, []) + [relation]
        neighbors[relation.col_type] = nbs_list
        self.adjacency_matrix[relation.row_type] = neighbors

    def add_relations_from(self, relations):
        """Add relations to the fusion graph.

        Parameters
        ----------
        relations : container of Relation-s
        """
        for relation in relations:
            self.add_relation(relation)

    def remove_relation(self, relation):
        """Remove a single relation from the fusion graph.

        Parameters
        ----------
        relation :
        """
        self.adjacency_matrix[relation.row_type][relation.col_type].remove(relation)
        self.relations.pop(relation)
        if relation.name:
            self._name2relation.pop(relation.name, None)
        if not self.adjacency_matrix[relation.row_type][relation.col_type]:
            self.adjacency_matrix[relation.row_type].pop(relation.col_type, None)
        if not list(self.in_neighbors(relation.row_type)) and \
                not list(self.out_neighbors(relation.row_type)):
            self.remove_object_type(relation.row_type)
            if relation.row_type == relation.col_type:
                return
        if not list(self.in_neighbors(relation.col_type)) and \
                not list(self.out_neighbors(relation.col_type)):
            self.remove_object_type(relation.col_type)

    def remove_relations_from(self, relations):
        """Remove relations from the fusion graph.

        Parameters
        ----------
        relations : container of Relation-s
        """
        for relation in relations:
            self.remove_relation(relation)

    def remove_object_type(self, object_type):
        """Remove a single relation from the fusion graph.

        Parameters
        ----------
        object_type :
        """
        for relation in self.relations:
            if object_type in relation:
                self.remove_relation(relation)
        self.adjacency_matrix.pop(object_type, None)
        for obj_type in self.adjacency_matrix:
            self.adjacency_matrix[obj_type].pop(object_type, None)
        self._name2object_type.pop(object_type.name, None)
        self.object_types.pop(object_type)

    def remove_object_types_from(self, object_types):
        """Remove relations from the fusion graph.

        Parameters
        ----------
        object_types: container of ObjectType-s
        """
        for object_type in object_types:
            self.remove_object_type(object_type)

    def get_relation(self, name):
        """Return a relation matrix with a given name.

        Parameters
        ----------
        name : str
            Name of the relation
        """
        if name not in self._name2relation:
            raise DataFusionError("Relation name unknown")
        return self._name2relation[name]

    def get_relations(self, row_type, col_type):
        """Return an iterator for relation matrices between two types of objects.

        Parameters
        ----------
        row_type : ObjectType
        col_type : ObjectType

        Returns
        -------
        relation :  iterator
        """
        if row_type not in self.object_types or col_type not in self.object_types:
            raise DataFusionError("Object types are not recognized.")
        return iter(self.adjacency_matrix.get(row_type, {}).get(col_type, []))

    def get_object_type(self, name):
        """Return object type whose name is provided.

        Parameters
        ----------
        name : str
            Name of the object type
        """
        if name not in self._name2object_type:
            raise DataFusionError("Object type name unknown")
        return self._name2object_type[name]

    def get_names(self, object_type):
        """Get names of all possible object type row/column names.

        Parameters
        ----------
        object_type : ObjectType

        Returns
        -------
        List of names when they exist, None otherwise.
        """
        if isinstance(object_type, str):
            object_type = self.get_object_type(object_type)

        size = 0
        for rel in self.out_relations(object_type):
            if rel.row_names:
                return rel.row_names
            else:
                size = rel.data.shape[0]

        for rel in self.in_relations(object_type):
            if rel.col_names:
                return rel.col_names
            else:
                size = rel.data.shape[1]

        return [str(x) for x in range(size)]

    def get_metadata(self, object_type):
        """Get metadata for given object type.

        Parameters
        ----------
        object_type : ObjectType

        Returns
        -------
        Metadata (list of dicts)
        """
        if isinstance(object_type, str):
            object_type = self.get_object_type(object_type)

        metadata = [{} for x in self.get_names(object_type)]

        for rel in self.out_relations(object_type):
            if rel.row_metadata:
                for md1, md2 in zip(metadata, rel.row_metadata):
                    md1.update(md2)

        for rel in self.in_relations(object_type):
            if rel.col_metadata:
                for md1, md2 in zip(metadata, rel.col_metadata):
                    md1.update(md2)
        return metadata


    def out_relations(self, object_type):
        """Return an iterator for relations adjacent to the object type.

        Parameters
        ----------
        object_type : ObjectType

        Returns
        -------
        relation : iterator
        """
        if object_type not in self.object_types:
            raise DataFusionError("Object type not in the fusion graph.")
        for col_type in self.adjacency_matrix.get(object_type, {}):
            for relation in self.adjacency_matrix[object_type][col_type]:
                yield relation

    def in_relations(self, object_type):
        """Return an iterator for relations adjacent to the object type.

        Parameters
        ----------
        object_type : ObjectType

        Returns
        -------
        relation : iterator
        """
        if object_type not in self.object_types:
            raise DataFusionError("Object type not in the fusion graph.")
        for row_type in self.adjacency_matrix:
            for relation in self.adjacency_matrix[row_type].get(object_type, {}):
                yield relation

    def out_neighbors(self, object_type):
        """Return an iterator for object types adjacent to the object type.

        Parameters
        ----------
        object_type : ObjectType

        Returns
        -------
        relation : iterator
        """
        if object_type not in self.object_types:
            raise DataFusionError("Object type not in the fusion graph.")
        return iter(self.adjacency_matrix.get(object_type, {}).keys())

    def in_neighbors(self, object_type):
        """Return an iterator for object types adjacent to the object type.

        Parameters
        ----------
        object_type : ObjectType

        Returns
        -------
        relation : iterator
        """
        if object_type not in self.object_types:
            raise DataFusionError("Object type not in the fusion graph.")
        for row_type in self.adjacency_matrix.keys():
            if object_type in self.adjacency_matrix[row_type]:
                if len(self.adjacency_matrix[row_type][object_type]) > 0:
                    yield row_type

    def __str__(self):
        return "{}(Object types: {}, Relations: {})".format(
            self.__class__.__name__, len(self.object_types), len(self.relations))

    def __repr__(self):
        return "{}(Object types={}, Relations={})".format(
            self.__class__.__name__,
            repr(list(self.object_types.keys())),
            repr(list(self.relations.keys())))


class ObjectType(object):
    """Object type used for fusion.

    Attributes
    ----------
    name :
    rank :
    """
    def __init__(self, name, rank=5):
        self.name = name
        self.rank = rank

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.name == other.name

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return self.name

    def __repr__(self):
        return '{}("{}")'.format(self.__class__.__name__, self.name)


class Relation(object):
    """Relation used for data fusion.

    Attributes
    ----------
    data :
    row_type :
    col_type :
    name :
    row_names :
    col_names :
    row_metadata :
    col_metadata :
    preprocessor :
    postprocessor :
    """
    def __init__(self, data, row_type, col_type, name='',
                 row_names=None, col_names=None,
                 row_metadata=None, col_metadata=None,
                 preprocessor=None, postprocessor=None, **kwargs):
        self.__dict__.update(locals())
        self.__dict__.update(kwargs)
        self.__dict__.pop('kwargs', None)
        self.__dict__.pop('self', None)
        self._id = name or uuid()

    def __contains__(self, obj_type):
        return obj_type == self.row_type or obj_type == self.col_type

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self._id == other._id

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return self.__repr__(str)

    def __repr__(self, repr=repr):
        return "{}({} {} {})".format(
            self.__class__.__name__,
            repr(self.row_type),
            ('"%s"' % self.name) if self.name else "→",
            repr(self.col_type))
